/*
 * AppBooster.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core;

import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.InfoHook;
import edu.utah.janos.nodeos.Status;

import edu.utah.janos.io.Log;

import edu.utah.janosvm.kit.comm.commspace.CommSpaceHandle;
import edu.utah.janosvm.kit.comm.commspace.CommQueuePuller;
import edu.utah.janosvm.kit.comm.commspace.CommSpaceElement;

import ants.util.InvalidKeyArgsException;

import java.util.Vector;

/**
 * Cannot be made public.
 * XXX: well, at least it can't be instantiated by the unwashed masses.
 * <p>
 * Applications do <emph>not</emph> have controls on their names space
 * usage installed.  They can access any name in the local namespace.
 *
 * CommSpaceHandle should point to a valid CSH for
 * the Booster class, plus:
 *   [2] this class will write the App port (integer) to this offset in the CSH.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class AppBooster
    extends Booster
{
    /**
     * Must have the appropriate key to create an AppBooster.
     * Only the core can actually create Keys, so only trusted
     * code can create AppBoosters...
     *
     * @exception SecurityException if the key is inappropriate
     * @exception NullPointerException if the key is null
     */
    public AppBooster(edu.utah.janos.nodeos.ClassInstantiationKey key)
	throws SecurityException
    {
	key.checkEquals(AppBooster.class);
    }
    
    protected void run(final Domain domain,
		       String[] args,
		       CommSpaceHandle commSpaceHandle)
    {
	BoosterArgs ba = new BoosterArgs(commSpaceHandle);
	final Application app;
	final Class appClass; 
	final Node localNode;
	
	Assert.require(domain);
	Assert.require(args);
	Assert.require(args.length >= 1);
	Assert.require(args[0]);
	
	domain.setName("App " +args[0]);
	
	// setupLocalNode was inherited from Booster
	localNode = setupLocalNode(domain, ba);
	
	try
	{
	    int lpc;
	    
	    // Add ants packages to the set of system packages.
	    for(lpc = 0; lpc < PrimordialNode.antsPackages.length; lpc++)
	    {
		Domain.currentDomain().addSystemPackage(PrimordialNode.
						    antsPackages[lpc]);
	    }
	    // Load the class and create initial instance
	    appClass = domain.loadClass(args[0]);
	    app = (Application)(appClass.newInstance()); 
	    
	    // Configure the app
	    app.setNode(localNode);
	    app.setName(args[0]);
	    app.setArgs(args, 1);
	    
	    // Install the app in the global application table
	    CommQueuePuller puller = app.getPacketQueue();
	    int port = app.getPort();
	    
	    /* Pass our queue to the primordial domain */
	    ba.setDomainMsgQueue(puller);
	    /* Ask the primordial domain for this port number */
	    ba.setAppPort(port);
	    
	    /* Wait for a reply from the primordial domain */
	    int realPort = puller.waitToTake().getInt();
	    
	    if( realPort == 0 )
	    {
		/* Port is busy */
		throw new AttachException("Port "
					  + port
					  + " is busy.");
	    }
	    else if( port == 0 )
	    {
		/* We asked for any port so set the real one */
		app.setPort(realPort);
	    }
	    /* Attach the application to the local node */
	    localNode.setPrincipal(app.getPrincipal());
	    localNode.attachApplication(app);
	    
	    Trace.trace(Trace.ANTS_APP, "App " +args[0]+ " attached at port " +app.getPort());
	    
	    // Start the handler for incoming packets
	    domain.threadStart(new Runnable() {
		public void run() {
		    /* XXX create a separate class AppPacketProcessor */
		    processPacket(domain, app, localNode);
		}}, "Packet Processor for " +args[0]);
	    
	    // Start the app
	    app.start();
	    
	    // Add the app to the application information
	    // XXX AppsInfo.add(args,app.getPort());
	    
	}
	catch (InvalidKeyArgsException ikae)
	{
	    localNode.log(Log.ERR, "AppBooster: Bad arguments passed to app " +args[0]+ ": ");
	    localNode.log(Log.ERR, ikae);
	    localNode.shutdown();
	}
	catch (AttachException ae)
	{
	    localNode.log(Log.ERR, "AppBooster: Cannot attach app " +args[0]+ ": ");
	    localNode.log(Log.ERR, ae);
	    localNode.shutdown();
	}
	catch (Exception e)
	{
	    localNode.log(Log.ERR, "AppBooster: Exception starting app " +args[0]+ ": ");
	    localNode.logStackTrace(Log.ERR, e);
	    localNode.shutdown();
	}
    }
    
    /**
     * Delivers packets from a protocol domain to an application.
     */
    private void processPacket(Domain domain, Application app, Node node)
    {
	CommQueuePuller queue = app.getPacketQueue();
	
	/*
	 * Loop processing packets from protocol domains.
	 * <p>
	 * XXX Looping on whether the domain was destroyed or not
	 * doesn't really work since we'll be stuck elsewhere and
	 * won't notice unless we get a packet...  Oh well, this
	 * should be taken care of by other stuff.
	 */
	while(!domain.destroyed())
	{
	    /* Grab the next capsule from our CommQueue */
	    Capsule cap = Capsule.dequeueCapsule(node, queue);
	    
	    /* Pass to the app for processing */
	    app.receive(cap);
	}
    }
    
    /**
     * InfoHook class that reports on the currently running
     * applications.
     */
    /*
      private static class AppsInfo
      extends InfoHook
      {
      static Vector apps = new Vector();
      
      private AppsInfo()
      {
      super("Apps");
      }
      
      public String[] getInfo()
      {
      String[] retval=new String[apps.size()];
      int i;
      for (i=0;i<apps.size();i++)
      retval[i]=(String)apps.elementAt(i);
      return retval;
      }
      
      public static void add(String[] args, int port)
      {
      String appinfo=new String();
      int i;
      for (i=0;i<args.length;i++)
      appinfo+=args[i]+" ";
      appinfo+="[port "+port+"]";
      apps.addElement(appinfo);
      }
      }
      
      static
      {
      Status.addInfoHook(new AppsInfo());
      }
    */
}
