/* 
 * Application.java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * Copyright (c) 2001 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package ants.core;

import edu.utah.janosvm.kit.comm.commspace.CommQueuePuller;
import edu.utah.janosvm.kit.comm.commspace.CommQueuePusher;
import edu.utah.janosvm.kit.comm.commspace.CommSpaceElement;

import ants.util.KeyArgs;
import ants.util.InvalidKeyArgsException;

import ants.core.security.Principal;
import ants.core.security.SecurityDefaults;

/**
 * Abstract node application.  An Application is an end-point for capsules.
 * Capsules are sent and received via the Node associated with the Application.
 * <p>
 * Applications understand the '-port' KeyArg for setting their port address
 * on the Node.  If no '-port' is provided an available port will be found
 * and allocated to the Application.  getPort() will retrieve the port
 * number.  
 * <p>
 * In addition, applications understand the '-principal' KeyArg for
 * setting the security principal.  In the abscence of this keyarg,
 * the principal defaults to SecurityDefaults.LOCAL_USERS. Some common
 * principles are described in SecurityDefaults.
 *
 * Applications are started after the Node has completed initialization.
 *
 * @see ants.core.security.SecurityDefaults
 * @author David Wetherall
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public abstract class Application
    extends Entity
{
    /**
     * Number of times to check for a valid address before giving up.
     *
     * @see #getNodeAddress()
     */
    private static final int GET_ADDRESS_MAX_TRIES = 10;

    /**
     * Amount of time (in millis) to wait between attempts to get
     * a valid address.
     *
     * @see #getNodeAddress()
     */
    private static final int GET_ADDRESS_WAIT_TIME = 500;

    /**
     * A queue to accept packets for the application.  The protocol for use
     * is to send the MID, followed by the packet bytes.
     */
    private final CommQueuePuller packetQueue = new CommQueuePuller();
    
    /**
     * Default KeyArgs for an application.  Sets port to 0.
     */
    public final static String[] defaults = { "-port", "0" };
    
    /**
     * The Node associated with this application.  All Capsules
     * are sent via and recieved from this Node.
     *
     * @see #setNode(Node)
     * @see #getNode()
     */
    private Node node;
    
    /**
     * The port for this application.  This is the identifier by
     * which remote nodes can send capsules to this application:
     * the address determines the Node, and the port determines
     * the application.
     * <p>
     * Set by the '-port' KeyArg.  Defaults to 0.
     * <p>
     * Is read-only after being set by '-port' (the read-only
     * semantics are required by Node.detachApplication()).
     * 
     * @see Node#attachApplication(Application)
     */
    private short port = 0;
    
    /**
     * The security principal of this object: used to make access
     * control decisions
     */
    private Principal principal;
    
    /**
     * The default TTL for packets sent by this application.
     */
    private final static int DEFAULT_TTL = 64;
    
    
    /**
     * The default resource limit for Capsules sent by
     * this application.
     *
     * @see #setDefaultTTL(int)
     * @see #getDefaultTTL()
     */
    private int ttl;
    
    public int getDefaultTTL()
    {
	return ttl;
    }
    
    public void setDefaultTTL(int ttl)
    {
	Assert.assert(ttl > 0);
	this.ttl = ttl;
    }
    
    /**
     * @deprecated use getDefaultTTL()
     */
    public int getDefaultResources()
    {
	return getDefaultTTL();
    }
    
    /**
     * @deprecated use setDefaultTTL(int)
     */
    public void setDefaultResources(int r)
    {
	setDefaultTTL(r);
    }
    
    /**
     * Returns the security principal.
     *
     * @see AppBooster#run(Domain,String[],CommSpaceHandle)
     */
    public final Principal getPrincipal() {
	return principal;
    }
    
    void setPrincipal(Principal p ) {
	this.principal = p;
    }
    
    
    /**
     * Get the integer port identifer for this Application.  Each
     * application on a node has a unique "port" identifier which
     * capsules use to direct themselves to particular applications
     * on a node.
     */
    public short getPort()
    {
	return port;
    }
    
    /**
     * Set this application's port.  Only
     * invoked by Node.attachApplication()
     * if this app didn't request a specific
     * port number.
     *
     * @see Node#attachApplication(Application)
     */
    void setPort(int p)
    {
	Assert.assert(port == 0);
	Assert.assert(p < (64 * 1024));
	port = (short)p;
    }
    
    /**
     * Return the Node interface customized for this application.
     */
    public Node thisNode() {
	return node; 
    }
    
    /**
     * Redundant accessor method
     *
     * @see #thisNode()
     */
    public Node getNode () {
	return node;
    }
    
    /**
     * Assign the node interface to this application.  Done
     * only in AppBooster.run().
     *
     * @see AppBooster#run(Domain,String[],CommSpaceHandle)
     */
    void setNode(Node node)
    {
	this.node = node;
    }

    /**
     * Return logical ANTS address of the node hosting this
     * application.  Unlike <code>Node.getAddress()</code>
     * which may return 0 (if the Node's address is not yet defined),
     * this version will block for a while until it gets a valid
     * address, or if it gets tired of waiting, it will throw
     * a NodeHasNoAddressError.
     *
     * @see Node#getAddress()
     */
    public int getNodeAddress()
    {
	Assert.require(this.node);

	int addr = this.node.getAddress();
	int tries = 0;

	while ((addr == 0)
	       && (tries < GET_ADDRESS_MAX_TRIES))
	{
	    try
	    {
		this.node.sleep(GET_ADDRESS_WAIT_TIME);
	    }
	    catch (InterruptedException ie)
	    {
		// oh well, just pretend we slept the whole time...
	    }
	    addr = this.node.getAddress();
	    tries++;
	    Trace.trace(-1, "next try getNodeAddress(): addr=" +addr+ "; tries=" +tries);
	}

	if (addr == 0)
	    throw new NodeHasNoAddressError("Node still has no address.");

	return addr;
    }
    
    /**
     * Send the given capsule.  
     * <p>
     * The capsule is sent with the default TTL for this application.
     *
     * @param cap the Capsule to send.  XXX state of Xdr?
     * @see #send(Capsule,int)
     * @see #ttl
     */
    public boolean send(Capsule cap)
    {
	return this.send(cap, this.ttl);
    }
    
    /**
     * Send the given capsule with the given resource limit.
     * <p>
     * The capsule is primed (that is, the 'from address' and ttl
     * are set), and then marshalled and sent to the protocol
     * domain that is responsible for the Capsule.
     * <p>
     * XXX this shouldn't return true/false.  It should succeed
     * or throw an exception.
     *
     * @param ttl the TTL to give the Capsule.
     * @see #ttl
     * @see Capsule#prime(int,int)
     * @see Node#send(Capsule)
     * @see PrimordialNode#getProtocolQueue()
     */
    public boolean send(Capsule cap, int ttl)
    {
	CommSpaceElement msg[] = { null, null, null, null };
	CommQueuePuller replyQueue;
	CommQueuePusher queue = null;
	boolean done = false;
	int sent = 0;
	
	/*
	 * XXX This kinda sucks, we need to return something so we
	 * have to setup a queue for getting the reply.
	 *
	 * XXX we don't get exceptions across this boundary, either.
	 */
	replyQueue = new CommQueuePuller();
	cap.prime(this.getNodeAddress(), ttl); // privileged prime()
	queue = PrimordialNode.getProtocolQueue(TypeID.wrap(cap.pid()));
	if( queue != null )
	{
	    msg[0] = new CommSpaceElement(ProtocolBooster.APP_CAPSULE);
	    cap.enqueueCapsule(msg, 1);
	    msg[3] = new CommSpaceElement(new CommQueuePusher(replyQueue));
	    queue.enqueue(msg);
	    while(!done)
	    {
		try
		{
		    sent = replyQueue.waitToTake().getInt();
		    done = true;
		}
		catch(InterruptedException e)
		{
		}
	    }
	} else {
		Trace.trace(Trace.ANTS_OUTGOING,"Application.java: queue==null");
	}
	return (sent != 0);
    }
    
    /**
     * Callback for packets that are delivered to this application.
     * This does nothing, so the default behavior for the application
     * is to drop delivered packets on the floor.
     *
     * @param cap the Capsule that got delivered.  XXX state of Xdr?
     */
    public void receive(Capsule cap)
    {
    }
    
    /**
     * Start this Application rolling.  This is invoked once by the
     * Node after it has been initialized.
     * <p>
     * Note that implementations of start should just set things up.  If
     * the application wants to do things like send packets, this
     * function should start a new thread to do that.  Creating a GUI
     * and associating callbacks is a good thing to do here.
     * <p>
     * XXX mostly stuff that could just be done in the no-arg 
     * constructor, no?
     */
    public abstract void start()
	throws Exception;
    
    /**
     * Parse command line arguments for this application.
     * Understands the -port option and the -principal option.
     * <p>
     * It is guaranteed that this method will be invoked
     * before the application starts running (i.e. shortly
     * after the constructor is invoked).
     * <p>
     * XXX A security issue related to this method is that a
     * mallicious application could change its parameters by
     * invoking this method after startup.  For example, the
     * application could change its principal to "Administrator".
     * To avoid any pitfalls, the local copies of the argument
     * variables should be ignored once control is given over to
     * the application's start method
     *
     * @see #port
     * @see KeyArgs
     */
    public void setArgs(KeyArgs k) 
	throws InvalidKeyArgsException
    {
	k.merge(defaults);
	
	for (int i = 0; i < k.length(); i++)
	{
	    if (k.key(i).equals("-port"))
	    {
		port = (short)Integer.parseInt(k.arg(i));
		k.strike(i);
	    }
	    else if (k.key(i).equals("-principal"))
	    {
		String who = k.arg(i);
		this.principal = new Principal (who);
		k.strike(i);
	    }
	}
	
	k.noMore();
    }
    
    /**
     * Return the packet queue for this application
     */
    final CommQueuePuller getPacketQueue()
    {
	return this.packetQueue;
    }
    
    /**
     * All applications must have a no-arg constructor as the
     * Node invokes Class.newInstance() to create the application.
     */
    public Application()
    {
	principal = SecurityDefaults.LOCAL_USERS;
	setDefaultTTL(DEFAULT_TTL);
    }
    
    public String toString()
    {
	return this.getClass().getName()
	    +"[node=" + this.node
	    + "; ttl=" + this.ttl
	    + "; port=" + this.port
	    + "]";
    }
}
