/*
 * BoosterArgs.java
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core;

import edu.utah.janosvm.kit.comm.commspace.CommSpaceHandle;
import edu.utah.janosvm.kit.comm.commspace.CommQueuePusher;
import edu.utah.janosvm.kit.comm.commspace.CommQueuePuller;
import edu.utah.janosvm.kit.comm.commspace.CommSpaceElement;
import edu.utah.janosvm.kit.comm.commspace.CommHashtableHandle;

/**
 * The BoosterArgs class encapsulates the various arguments we need
 * to pass down to any new domains created.
 *
 * @see Booster
 * @see AppBooster
 * @ see edu.utah.janos.nodeos.CommSpacePusher
 * @ see edu.utah.janos.nodeos.CommSpacePuller
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
class BoosterArgs
{
    /**
     * IN: Protocol ID for this protocol domain, if it is one
     */
    private final static int PID		= 0;
    /**
     * IN: MID to space map needed for the UnknownMidPackets handler
     */
    private final static int MID_TO_SPACE_MAP	= 1;
    /**
     * OUT: Message queue used to communicate with the domain
     */
    private final static int DOMAIN_MSG_QUEUE	= 2;
    /**
     * IN/OUT: Used to negotiate the port for an application
     */
    private final static int APP_PORT		= 3;
    /**
     * Maximum number of slots in the CommSpace
     */
    private final static int ARG_MAX		= 4;
    
    /**
     * CommSpace used to communicate with the primordial domain
     */
    private CommSpaceHandle commSpace;
    
    /**
     * Construct a BoosterArgs object that will take from the given
     * CommSpace.
     *
     * @param csh The CommSpace to read from.
     */
    BoosterArgs(CommSpaceHandle csh)
    {
	this.commSpace = csh;
    }
    
    /**
     * Construct a BoosterArgs that can be used for a Booster in a new
     * domain.
     *
     * @param pid The Protocol ID if its a protocol domain.
     * @param midToSpaceMap The MID to Space map for the UnknowMidHandler
     */
    BoosterArgs(TypeID pid,
		CommSpaceHandle midToSpaceMap)
    {
	this.commSpace = CommSpaceHandle.newCommSpace(ARG_MAX);
	if( pid != null )
	{
	    this.commSpace.write(PID,
				 new CommSpaceElement(pid.
						      unwrap()));
	}
	this.commSpace.write(MID_TO_SPACE_MAP,
			     new CommSpaceElement(midToSpaceMap));
    }
    
    /**
     * Get the CommSpace that the BoosterArgs values are held in.
     */
    CommSpaceHandle getCommSpaceHandle()
    {
	return this.commSpace;
    }
    
    TypeID getPID()
    {
	return TypeID.wrap(this.commSpace.take(PID).getByteArray());
    }
    
    UnknownMidPackets getUnknownMidPackets()
    {
	return new UnknownMidPackets(this.commSpace.
				     take(MID_TO_SPACE_MAP).
				     getCommSpaceHandle());
    }
    
    void setDomainMsgQueue(CommQueuePuller puller)
    {
	this.commSpace.write(
			     DOMAIN_MSG_QUEUE,
			     new CommSpaceElement(new CommQueuePusher(puller)));
    }
    
    CommQueuePusher waitForDomainMsgQueue()
    {
	try
	{
	    return this.commSpace.waitToTake(DOMAIN_MSG_QUEUE).
		getCommQueue();
	}
	catch(InterruptedException ie)
	{
	    return null;
	}
    }
    
    void setAppPort(int port)
    {
	this.commSpace.write(APP_PORT, new CommSpaceElement(port));
    }
    
    int waitForAppPort()
    {
	try
	{
	    return this.commSpace.waitToTake(APP_PORT).getInt();
	}
	catch(InterruptedException ie)
	{
	    return 0;
	}
    }
}
