/*
 * CapsuleChannel.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core;

import edu.utah.janos.nodeos.BufferHandle;
import edu.utah.janos.nodeos.DeliverFunc;
import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.InChannel;
import edu.utah.janos.nodeos.ProtocolSpecFormatException;
import edu.utah.janos.nodeos.PacketAddressFormatException;

import edu.utah.janos.io.Log;

import java.io.IOException;

/**
 * The handler for incoming ANTS packets.  They get decoded into
 * Capsules and evaluted here.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
final class CapsuleChannel
    implements DeliverFunc
{
    
    private static final int DEFAULT_BUFFER_COUNT = 20;
    
    Class cl;
    InChannel inChannel;
    private ANTSDemultiplexKey classifier;
    
    /**
     * Construct a receive channel for cc Capsules.  Implicitly
     * constructs an ANTSDemultiplexKey to match the right packets.
     */
    CapsuleChannel(CapsuleClass cc)
    {
	this(cc.cl,
	     new ANTSDemultiplexKey(Domain.currentDomain(), cc.methodID));
    }
    
    CapsuleChannel(Class cl, ANTSDemultiplexKey classifier)
    {
	Domain domain = Domain.currentDomain();
	
	Assert.assert(cl);
	Assert.assert(classifier);
	
	this.cl = cl;
	this.classifier = classifier;
	
	Trace.trace(Trace.ANTS_CAPSULECHANNEL, "new capsule channel: class=" +cl+ "; classifier=" +classifier);
	try
	{
	    inChannel = InChannel.create(domain,
					 classifier,
					 PrimordialNode.getReceiveProtocolSpec(),
					 PrimordialNode.getReceiveAddressSpec(),
					 this);
	    inChannel.addBuffers(DEFAULT_BUFFER_COUNT);
	}
	catch(PacketAddressFormatException e)
	{
	    Node.abort(e, "Bad physical address spec:");
	}
	catch(ProtocolSpecFormatException e)
	{
	    Node.abort(e, "Bad physical protocol spec:");
	}
	catch(IOException e)
	{
	    Node.abort(e, "IOException when creating in channel:");
	}
    }
    
    /**
     * Handler for packets delivered to this Channel.
     * <p>
     * Receives a BufferHandle and decodes the packet
     * into a Capsule.
     * <p>
     * Assumes packet[0] is the first byte of the ANEP
     * header...
     * <p>
     * XXX this is atrocious code.	Rewrite it now.
     */
    public void receive(BufferHandle packet, int payloadStart)
	throws IOException
    {
	Assert.assert(packet);
	Assert.assert(this.cl);
	Assert.assert(this.classifier);
	
	// XXX
	//
	// XXX This is old code from before I knew what was
	// going on.  Don't trust it.  :) OTOH, it could be
	// right...
	// 
	// XXX
	
	PrimordialNode.getStats().receivedPacket();
	
	// XXX Xdr should wrap a BufferHandle better than this!
	byte[] buf = new byte[packet.getLength()];
	packet.getBytes(0, buf, 0, buf.length);
	
	// Recycle the BufferHandle
	inChannel.addBuffer(packet); 
	
	Xdr xdr = new Xdr(buf);
	if(Trace.ENABLED) Trace.trace(Trace.ANTS_CAPSULECHANNEL, "buf.length = " +buf.length+ ", payloadStart = " +payloadStart);
	
	Capsule c = null;
	try
	{
	    if(Trace.ENABLED) Trace.trace(Trace.ANTS_CAPSULECHANNEL, "capsule channel instantiating " +cl);
	    c = (Capsule)cl.newInstance();
	    // c.mid = classifier.getMID();
	}
	catch (Throwable th)
	{
	    Node.abort(th, "receive() blew up creating a new instance of " +cl.getName());
	}
	
	// increment xdr through ANEP crap
	// also skip the ANTS typeid
	
	xdr.skip(classifier.whereDoesTheCapsuleDecodingActuallyStart(payloadStart));
	
	// Decode the xdr into the Capsule
	c.decodeWith(xdr); // XXX exceptions??
	
	// Evaluate the new Capsule
	Node node = Node.currentNode();
	try
	{
	    boolean sentIt;
	    sentIt = c.evaluate(node);
	}
	catch (Throwable th)
	{
	    // XXX use Trace/Log/etc...
	    node.log(Log.ERR, "Capsule " +c+ " exploded: " +th.getMessage());
	    node.logStackTrace(Log.ERR, th);
	}
    }
    
    public String toString()
    {
	return "CapsuleChannel[cl=" + this.cl.getName()
	    + "; inChannel= XX"
	    // + this.inChannel // XXX circular toString() references
	    + "; classifier=" + this.classifier
	    + "]";
    }
}
