/* 
 * CodeGroup.java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package ants.core;

import ants.util.MD5;

import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.InChannel;
import edu.utah.janos.nodeos.InChannelHandle;

import java.util.Vector;

/**
 * A record collecting information about methods into a related group.
 * The CodeGroup is the unit of code forwarding.
 * <p>
 * This class is only used internally in ANTS and is part of the
 * dynamic code loading infrastructure.
 * <p>
 * The fields of CodeGroup have package access since a CodeGroup may
 * be in many different states of completion during code loading and
 * protocol registration.
 * <p>
 * A protocol is composed of several CodeGroups.
 *
 * @see Protocol
 * @author David Wetherall
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
class CodeGroup
{
    /**
     * The CapsuleClasses comprising the CodeGroup.
     *
     * XXX make into an array?
     */
    final Vector classes = new Vector();
    
    final TypeID groupID;
    
    /**
     * groupIDs may or may not be necessary.  Currently, it's only
     * used to fill in the group ID list in a DLResponse capsule.
     * groupIDs holds the groupID of every group in the protocol.
     * This may be useful for checking that a group really is a
     * member of the protocol: a group belongs to a protocol only
     * if it is in the group ID list and the group ID list hashes
     * to the protocol ID.  We may not need this, though.
     * <p>
     * The groupIDs is fixed at instantiation time.  The contents
     * of this array should not be modified by anyone.
     */
    final TypeID[] groupIDs;
    
    /**
     * ProtocolID is computed from the Vector of GroupIDs.  This
     * is the only link from a CodeGroup up to its Protocol.
     * <p>
     * It may seem wasteful to store the ProtocolID in each CodeGroup
     * in a Protocol, but remember, Protocol objects only exist
     * at application nodes.  Shipping the Protocol object just to
     * give internal nodes a handle on the ProtocolID would be
     * even more wasteful.
     */
    final TypeID protocolID;
    
    /**
     * The number of chunks of size CHUNK_SIZE that are required to
     * ship this CodeGroup.
     * <p>
     * There is a hard limit of 32 chunks in a CodeGroup.  (Because the
     * set of chunks in a CodeGroup a node has loaded must fit in a
     * 32-bit mask.)
     * <p>
     * Each CapsuleClass in the CodeGroup will require a set
     * number of chunks to transport.  Sadly, chunks for distinct
     * CapsuleClasses are not compressed together, i.e., two
     * CapsuleClasses of CHUNK_SIZE+4 will require 4 chunks to
     * ship.
     *
     * @see #CHUNK_SIZE
     * @see #availChunkMask
     */
    int chunkCount = 0;
    
    /**
     * A bit vector indicating which code chunks are available on
     * this node.  Extra bits are 1 (i.e., if a CodeGroup has four
     * chunks, and all four are loaded and available all of the
     * bits in the availChunkMask will be set.
     * <p>
     * XXX add availChunkMask() function.
     * XXX make this a long?!
     */
    int availChunkMask;
    
    /**
     * Flag if this class is a builtin class.
     */
    boolean builtin = false;
    
    /**
     * Code to be shipped over the net is broken into
     * chunks no bigger than this value.
     * <p>
     * XXX ideally this value should be a function of the
     * outgoing link's MTU.
     *
     * @see #codeChunkCount(byte[])
     */
    final static int CHUNK_SIZE = 1024;
    
    
    final boolean builtin()
    {
	return builtin;
    }
    
    /**
     * Compute a mask for chunk i.  Fills in all unused chunk
     * fields with 1.  Thus the result can be read as a bit
     * mask with 0's indicating which chunks are part of the
     * CodeGroup except chunk i is set.
     *
     * @param i the chunk index
     * @return a bit-vector indicating which chunks are available.
     */
    final int chunkMask(int i)
    {
	Assert.assert(i < chunkCount);
	return ((1<<i) | ((0xffffffff)<<this.chunkCount));
    }
    
    /**
     * Determine the number of chunks the given code array will
     * be broken into.
     * <p>
     * XXX rename to chunkCountForCode()
     *
     * @return the number chunks the given code array will be broken into.
     */
    static int codeChunkCount(byte[] code)
    {
	if (code == null)
	{
	    // Builtin capusle types have a code size of 0
	    // because they will never be shipped around.
	    // XXX Builtin's probably shouldn't get here.
	    return 0;
	}
	
	return (code.length + CHUNK_SIZE - 1) / CHUNK_SIZE;
    }
    
    /**
     * Register a CodeGroup with the current domain.  Essentially this just
     * means that objects created from group classes can be instantiated
     * and sent, but dispatching is taken care of by another domain.
     *
     * @param node The node to register this group with
     */
    final void register(Node node)
	throws ClassNotFoundException
    {
	Domain domain = Domain.currentDomain();
	
	Trace.trace(Trace.ANTS_DYNLOAD, "Registering " +this+ " in " +domain);
	
	// Registration and verification of all classes in this group

	// First we register them all
	for (int i = 0; i < classes.size(); i++) 
	{
	    CapsuleClass cc;
	    cc = (CapsuleClass)classes.elementAt(i);
	    
	    // verify code, group and protocol match method
	    if (!builtin() && !cc.checkMethodID()) 
	    {
		Node.abort("Bad protocol identifier in CodeGroup.register()");
	    }
	    
	    // Install the code in the Domain's classloader.
	    if (!builtin())
		domain.addClass(cc.name(), cc.code());
	}

	// Second we cause them to load
	for (int i = 0; i < classes.size(); i++)
	{
	    CapsuleClass cc;
	    cc = (CapsuleClass)classes.elementAt(i);
	    
	    try
	    {
		cc.cl = domain.loadClass(cc.name());
		Assert.assert(cc.cl != null);
	    }
	    catch (Throwable th)
	    {
		// Note, we convert bogus JVM exceptions like
		// NoClassDefFoundError in ClassNotFoundException
		Trace.trace(Trace.ANTS_DYNLOAD, "Couldn't load class " +cc.name()+ ": " +th.getMessage());
		throw new ClassNotFoundException("domain.loadClass('" +cc.name()+ "') threw "
						 +th.getClass().getName()+
						 ": " +th.getMessage());
	    }

	    Trace.trace(Trace.ANTS_DYNLOAD, "Added class " +cc.name()+ " to " +this);

	    /*
	     * Force static initializers of builtin capsule types
	     * to run in current context (the root domain context).
	     *
	     * XXX this is really lame.  But, if we don't do this
	     * then the "first use" of a class may be in a sub-domain
	     * when that sub-domain subclasses a builtin capsule
	     * (e.g., PingCapsule subclassing DataCaspule).  The
	     * problem is that static intiailzers on DataCapsule
	     * (in particular findMID()) will be run in the sub-domain
	     * context.  They will fail in that context as they are
	     * run against the sub-domains "Node.currentNode()" instead
	     * of the builtin Node.currentNode()...
	     */
	    if (builtin())
	    {
		try
		{
		    cc.cl.newInstance();
		}
		catch (Exception e)
		{
		    Node.abort(e, "Exception in newInstance hack.  Should never happen.");
		}
	    }
	}
    }
    
    /**
     * Activate this CodeGroup. All classes in the CodeGroup are registered
     * with the domain, and instances of all the various Capsules classes
     * are created.  If the CodeGroup is "actually active", then InChannels
     * are registered for DLRequests and MIDIDRequests for the protocol.
     * Any existing packets queued up for this protocol are forwarded to
     * the new InChannels.
     * <p>
     * Invoked by Protocol.activate() and DLResponseCapsule.evaluate().
     *
     * @param node the Node to get the waiting packets off of
     * @param actuallyActive flag to indicate if this CodeGroup should have InChannels created for it.
     */
    final void activate(Node node, boolean actuallyActive)
	throws ClassNotFoundException
    {
	// start with all chunks present
	availChunkMask = 0xffffffff;
	
	Domain domain = Domain.currentDomain();
	
	Trace.trace(Trace.ANTS_DYNLOAD, "Activating " +this+ " in " +domain);
	
	// XXX may throw ClassNotFoundException or NoClassDefFoundError
	this.register(node);
	
	// XXX add all the non-Capsule classes to the Domain...
	
	// All the bytecode is in place, so we can create the
	// class objects and in-channels:
	if( actuallyActive )
	{
	    for (int i = 0; i < classes.size(); i++) 
	    {
		CapsuleClass cc;
		cc = (CapsuleClass)classes.elementAt(i);

		// Create InChannel for the Capsule
		InChannel inChannel = new CapsuleChannel(cc).inChannel;
		
		// For code that can be moved, create the necessary
		// DL and MIDID request handlers:
		if (!builtin())
		{
		    // Note: this relies on a correct VM to
		    // make sure that the static MID fields don't
		    // get initialized if builtin()
		    new CapsuleChannel(MIDIDRequestCapsule.class,
				       new CodeRequestDemultiplexKey(domain,
								     MIDIDRequestCapsule.MID,
								     cc.methodID.unwrap()));
		    
		    new CapsuleChannel(DLRequestCapsule.class,
				       new CodeRequestDemultiplexKey(domain,
								     DLRequestCapsule.MID,
								     cc.methodID.unwrap()));
		}
		
		// Forward queued up packets to the InChannel
		node.unknownMidPackets.setInChannel(cc.methodID.unwrap(),
						    new InChannelHandle(inChannel));
		/*
		 * At this point the new InChannel for this
		 * type of MID has been created and has all of
		 * the packets that have been queued up for it.
		 * It can proceed on its own now.
		 */
	    }
	    
	}
    }
    
    /**
     * Add a capsule to the code group.
     *
     * @exception FullCodeGroupException if the CodeGroup has too many chunks
     * (currently the limit is 32).
     */
    final void add(CapsuleClass cc)
	throws FullCodeGroupException
    {
	chunkCount += codeChunkCount(cc.code());
	
	// sanity check limit
	if (chunkCount > 32)
	    throw new FullCodeGroupException("CodeGroup "+ groupID +" too long.  (Already has 32 chunks).");
	
	classes.addElement(cc);
    }
    
    public String toString()
    {
	return "CodeGroup[gid=" + groupID
	    + "; ct=" + chunkCount
	    + "; cmask=" + Integer.toHexString(availChunkMask)
	    + "; " + (builtin ? "builtin" : "dynamic")
	    + "]";
    }
    
    /**
     * Create a new (empty) CodeGroup with
     * the given groupID and protocolID.
     *
     * @param groupID The group ID for this CodeGroup.
     */
    CodeGroup(TypeID groupID, TypeID[] groupIDs, TypeID protocolID, boolean builtin)
    {
	this.groupID = groupID;
	this.groupIDs = groupIDs;
	this.protocolID = protocolID;
	this.builtin = builtin;
	
	Node.currentNode().publishCodeGroup(this);
    }
}
