/* 
 * DataCapsule.java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package ants.core;

/**
 * Data capsule.  Data capsules simply forward themselves to their
 * destination.  Each DataCapsule contains source and destination
 * application ports (in addition to the source and destination
 * node addresses kept by the superclass).  Also, an arbitrary
 * ByteArray payload is attached to the capsule.
 * <p>
 * XXX This doesn't need to be a builtin capsule type.  Its convenient
 * and all, but its not critical.  Move it out to the apps/ directory.
 * (Well, the fact that its builtin means the ID is fixed, so this
 * is handy in a backwards-compatiblity way.)
 *
 * @author David Wetherall
 */
public class DataCapsule 
    extends Capsule 
{
    
    /**
     * WARNING: this is a builtin capsule, whose version number
     * must be manually kept in synch with  its functionality.
     *
     * @see BuiltinProtocol#addCapsule(java.lang.String,java.lang.String)
     */
    final public static String VERSION = "ANTS2.0 DataCapsule October 13, 2000";
    
    final private static byte[] MID = findMID("ants.core" + ".DataCapsule");
    final private static byte[] PID = findPID("ants.core" + ".DataCapsule");
    
    protected byte[] mid()
    {
	return MID;
    }
    
    protected byte[] pid()
    {
	return PID;
    }
    
    /**
     * Source application's port.
     * <p>
     * Not used by the DataCapsule, but its assumed that most users of
     * this type will want to know where the capsule came from.
     */
    public short spt;
    
    /**
     * Destination application's port.
     */
    public short dpt;
    
    /**
     * Data payload this capsule pulls through the network.
     */
    private ByteArray data;
    
    /**
     * Return length of this data capsule.
     *
     * @see Capsule#length()
     */
    public int length()
    {
	int s = Xdr.SHORT + // spt
	    Xdr.SHORT + // dpt
	    Xdr.SHORT + // array length
	    (data == null ? 0 : data.length());
	return super.length() + s;
    }
    
    public Xdr encode()
    {
	Xdr xdr = super.encode();
	xdr.PUT(spt);
	xdr.PUT(dpt);
	xdr.PUT(data);
	return xdr;
    }
    
    public Xdr decode()
    {
	Xdr xdr = super.decode();
	spt = xdr.SHORT();
	dpt = xdr.SHORT();
	data = xdr.BYTEARRAY();
	return xdr;
    }
    
    /**
     * Get the payload associated with this DataCapsule.
     */
    public ByteArray getData()
    {
	return data;
    }
    
    /**
     * Set the payload associated with this DataCapsule.
     */
    public void setData(ByteArray data)
    {
	this.data = data;
    }
    
    /**
     * Forwarding function for DataCapsules.  Forwards the capsule to
     * the appropriate application port if the capsule is on its
     * destination node, otherwise forwards the capsule on to its
     * destination node.
     */
    public boolean evaluate(Node n)
    {
	if (n.getAddress() ==  getDst())
	{
	    return n.deliverToApp(this, dpt);
	}
	else
	{
	    return n.routeForNode(this, getDst());
	}
    }  
    
    public DataCapsule()
    {
    }
    
    /**
     * Construct a new DataCapsule.
     *
     * @param s the source application's port
     * @param d the destination application's port
     * @param n the address of the destination node.
     * @param p the payload.
     */
    public DataCapsule(short s, short d, int n, ByteArray p)
    {
	spt = s;
	dpt = d;
	setDst(n);
	data = p;
    }
    
    public String toString()
    {
	StringBuffer buf = new StringBuffer(100);
	
	buf.append(super.toString());
	buf.append("+DataCapsule[spt=");
	buf.append(spt);
	buf.append("; dpt=");
	buf.append(dpt);
	buf.append("; data=");
	buf.append(data);
	buf.append("]");
	return buf.toString();
    }
}
