/* 
 * Entity.java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package ants.core;

import ants.util.KeyArgs;
import ants.util.InvalidKeyArgsException;

/**
 * An Entity is an abstract class from which all Node-resident
 * "things" are subclassed.  For example, applications, channels,
 * extensions and the node interface are all Entities.
 * <p>
 * Every entity has a name.
 * <p>
 * Every entity has a set of "arguments" it understands.  These
 * arguments are represented as string pairs: an argument name
 * and argument value.  Each subclass of entity can define
 * its own arguments and their default values.  At this level,
 * the only argument is '-log' (defaults to '0'.)
 *
 * @see ants.util.KeyArgs
 * @author David Wetherall
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public abstract class Entity
{
    
    /**
     * Default arguments that are handled at the Entity level.
     * <p>
     * All subclasses should follow this pattern: a <code>public final
     * static String[]</code> named <code>defaults</code> so that,
     * via reflection, usage information can be obtained in a generic
     * manner.
     * <p>
     * Subclasses should explicitly set the defaults to null.
     * 
     * @see #setArgs(ants.util.KeyArgs)
     */
    private final static String[] defaults = null;
    
    /**
     * Name of this entity.
     */
    private String name;
    
    /* --- Entity name interface --- */
    
    String getName()
    {
	return name;
    }
    
    void setName(String name)
    {
	this.name = name;
    }
    
    /* --- Argument-related interface --- */
    
    String usage()
    {
	Class top = this.getClass();
	Class base = Object.class;
	KeyArgs k = null;
	
	String us = top.getName() + ": ";
	while (top != base) {
	    try 
	    {
		if (k == null)
		    k = new KeyArgs();
		java.lang.reflect.Field f = top.getDeclaredField("defaults");
		k.merge((String[])f.get(top));
	    }
	    catch (InvalidKeyArgsException ikae)
	    {
				// skip
	    }
	    catch (NoSuchFieldException nfse)
	    {
				// skip this entry
	    }
	    catch (NullPointerException npe)
	    {
				// skip this entry
	    }
	    catch (IllegalArgumentException iae)
	    {
		Node.abort("don't have access to defaults field in " +top.toString());
	    }
	    catch (IllegalAccessException iae)
	    {
		Node.abort("don't have access to defaults field in " +top.toString());
	    }
	    catch (SecurityException se)
	    {
		Node.abort("don't have access to defaults field in " +top.toString());
	    }
	    top = top.getSuperclass();
	}
	return us + k.toString();
    }
    
    /**
     * Set this Entity's arguments to the given array of strings.  The
     * array must contain an even number of Strings (i.e., a whole
     * number of pairs).
     * <p>
     * Subclasses should override the setArgs(ants.util.KeyArgs)
     * version of this method if they want to handle their
     * own arguments.
     *
     * @param ka the array of String arguments.
     * @exception InvalidKeyArgsException if there are unused arguments
     */
    void setArgs(String[] ka) 
	throws InvalidKeyArgsException
    {
	setArgs(ka, 0);
    }
    
    /**
     * Set this Entity's arguments to the given array of strings.  The
     * array must contain an even number of Strings (i.e., a whole
     * number of pairs).  Start at index 'sk', ignoring previous
     * entries in the array.
     * <p>
     * Subclasses should override the setArgs(ants.utilKeyArgs)
     * version of this method if they want to handle their
     * own arguments.
     *
     * @param ka the array of String arguments.
     * @param sk the start key in the array.
     * @exception InvalidKeyArgsException if there are unused arguments
     */
    void setArgs(String[] ka, int sk)
	throws InvalidKeyArgsException
    { 
	setArgs(new KeyArgs(ka, sk));
    }
    
    /**
     * Set this Entity's arguments to the those specified in k.  Default
     * arguments are overriden if specified in k.
     * <p>
     * Since this is the last place that handles KeyArgs, if the there
     * are any remaining arguments, an exception is thrown.  Subclasses
     * of Entity must 'strike()' any KeyArgs they consume before
     * invoking this method.
     *
     * @param k the arguments to use.
     * @exception InvalidKeyArgsException if there are extra arguments in k
     * @see ants.util.KeyArgs#strike(int)
     */
    void setArgs(KeyArgs k) 
	throws InvalidKeyArgsException
    {
	k.merge(defaults);
	k.noMore();
    }
    
    /**
     * No-argument constructor for Entities created through
     * Class.newInstance().  Results in an unnamed entity.
     * <p>
     * Name must be set through setName();
     */
    Entity()
    {
    }
    
    /**
     * Construct a named Entity.
     */
    Entity(String name)
    {
	setName(name);
    }
    
    /*
     * --- OLD-STYLE Logging interface
     */
    
    /**
     * @deprecated Use the Janos I/O Logging levels instead.
     */
    protected int[] L = new int[10];
    
    /**
     * @deprecated Use node.log() for real logging.  Use TRACE macros for debugging.
     */
    protected void log(int level, String msg)
    {
	PrimordialNode.message(level, msg);
    }
    
    public String toString()
    {
	return "Entity[name="
	    + this.name
    + "]";
    }
}
