/* 
 * MIDIDRequestCapsule.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core;

import edu.utah.janos.io.Log;

import java.util.Vector;

/**
 * MID ID Request capsule.
 * <p>
 * The MIDIDRequestCapsule adds one field to the base Capsule:
 * <ul>
 *  <li> the typeid of the unknown MID (this.unknownMID)
 * </ul>
 *
 * <p> Pat said to make a note that this capsule could be a domain
 * control message for data packets.  It says, "woah, stop sending
 * data until the code is loaded".  But we don't take advantage of
 * this yet.
 *
 * @see DLProtocol
 * @see MIDIDResponseCapsule
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
final class MIDIDRequestCapsule
    extends Capsule 
{
    // WARNING: this is a builtin capsule, whose version number
    // must be manually kept in synch with  its functionality
    public final static String VERSION = "ANTS 2.0 MIDIDRequestCapsule October 13, 2000";
    
    final static byte[] MID = findMID("ants.core" + ".MIDIDRequestCapsule");
    private final static byte[] PID = findPID("ants.core" + ".MIDIDRequestCapsule");
    
    protected byte[] mid()
    {
	return MID;
    } 
    
    protected byte[] pid()
    {
	return PID;
    }
    
    /**
     * The TypeID of the unknown MID this Capsule represents.
     * This is set only by the constructor (for outgoing
     * MIDIDRequestCapsules) or by the decode() function (for
     * incoming MIDIDRequestCapsules).
     */
    private TypeID unknownMID;
    
    public int length()
    {
	return super.length() + Xdr.TYPEID;
    }
    
    public Xdr encode() 
    {
	Xdr xdr = super.encode();
	xdr.PUT(unknownMID);
	
	return xdr;
    }
    
    public Xdr decode() 
    {
	Xdr xdr = super.decode();
	unknownMID = xdr.TYPEID();
	
	return xdr;
    }
    
    /**
     * Evaluate a MIDIDRequestCapsule.  Based on the unknownMID stored
     * in this capsule, lookup the CapsuleClass on the local node.
     * If the CapsuleClass is not available drop the request on
     * the floor.  Otherwise, send a MIDIDResponseCapsule with the
     * Protocol identity of the MID.
     * <p>
     * If the MIDIDRequestCapsule is for a builtin type, just drop
     * the request.
     *
     * @see CodeGroup
     */
    public boolean evaluate(Node node)
    {
	boolean ret = true;
	
	if( node.getAddress() != getDst() )
	{
	    return node.routeForNode(this, getDst());
	}
	
	// Look the desired method up on the local node
	CapsuleClass cc = node.lookupCapsuleClass(unknownMID);
	Trace.trace(Trace.ANTS_DYNLOAD, "cc for mid " +unknownMID+ " -> " +cc);
	
	if ((cc == null) || (cc.cl == null))
	{
	    /*
	     * This can only happen if the cc is not here
	     * at all or just barely being loaded on this
	     * node.  Whomever asked for it is broken or
	     * trying to annoy me.
	     */
	    
	    // XXX these messages should back off exponentially.
	    node.log(Log.WARNING,
		     "MIDIDRequest: Ignoring CapsuleClass request from belligerent node "
		     +NodeAddress.toString(getSrc())+ ".  (cc=" +cc+ ")");
	    
	    return true;
	}
	
	// test to see if this is builtin and non-transferable
	if (cc.group.builtin())
	{
	    // XXX these messages should back off exponentially.
	    node.log(Log.WARNING,
		     "MIDIDRequest: Node " +NodeAddress.toString(getSrc())+
		     " wanted builtin capsule.  Dropping request.");
	    
	    return true;
	}
	
	Trace.trace(Trace.ANTS_DYNLOAD, "MIDIDRequest for " +unknownMID+ " from " +NodeAddress.toString(getSrc()));
	
	/*
	 * Send off one MIDIDResponseCapsule.
	 */
	Capsule dl = new MIDIDResponseCapsule(getPrevious(),
					      unknownMID,
					      cc.protocolID(),
					      cc.group.groupID,
					      cc.userID()); // XXX
	
	// send the packet with a TTL of 1 ...
	dl.prime(node.getAddress(), 1);
	
	return node.routeForNode(dl, getPrevious());
    }
    
    /**
     * Create a blank MIDIDRequestCapsule.
     * <p>
     * This constructor should only be invoked when
     * a MIDIDRequestCapsule is being received and decoded.
     */
    MIDIDRequestCapsule()
    {
    }
    
    /**
     * Create a MIDIDRequestCapsule that will request info on
     * the given unknown MID.
     * <p>
     * This constructor is only invoked by DLProtocolHandler.  // XXX which is being renamed.
     */
    MIDIDRequestCapsule(int dst, TypeID unknownMID)
    {
	setDst(dst);
	Assert.assert(unknownMID != null);
	this.unknownMID = unknownMID;
    }
    
    public String toString()
    {
	StringBuffer buf = new StringBuffer(100);
	
	buf.append(super.toString());
	buf.append("+MIDIDRequest[unknownMID=");
	buf.append(unknownMID);
	buf.append("]");
	return buf.toString();
    }
}
