/* 
 * MIDIDResponseCapsule.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core;

/**
 * MID ID Response Capule.
 * <p>
 * MIDIDResponseCapsules are only created by a MIDIDRequestCapsule
 * that has been delivered to a node that can satisfy the request.
 * MIDIDResponseCapsules contain information describing the MID.
 * <p>
 * A MIDIDResponseCapsule packet contains the following fields (on top of the
 * base Capsule fields):
 * <ul>
 *   <li> requestMID: MID that initiated the request
 *   <li> pid: PID associated with the MID.
 *   <li> userInfo: userInfo (see UserID).
 * </li>
 * 
 * @see DLProtocol
 * @see MIDIDRequestCapsule
 * @see CodeGroup
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
class MIDIDResponseCapsule
    extends Capsule 
{
    // WARNING: this is a builtin capsule, whose version number
    // must be manually kept in synch with  its functionality
    public final static String VERSION = "ANTS 2.0 MIDIDResponseCapsule October 13, 2000";
    
    private final static byte[] MID = findMID("ants.core" + ".MIDIDResponseCapsule");
    private final static byte[] PID = findPID("ants.core" + ".MIDIDResponseCapsule");
    
    protected byte[] mid()
    {
	return MID;
    }
    
    protected byte[] pid()
    {
	return PID;
    }
    
    /**
     * The MID that initiated the request.
     */
    private TypeID requestMID;
    
    /**
     * PID associated with the requestMID.
     */
    private TypeID pid;
    
    /**
     * GID associated with the requestMID.
     */
    private TypeID gid;
    
    /**
     * User information associated with the PID.
     */
    private UserID userInfo;
    
    public int length() 
    {
	return super.length() +
	    Xdr.TYPEID +		// groupID
	    Xdr.TYPEID +		// payloadID
	    Xdr.TYPEID +		// protocolID
	    userInfo.length();
    }
    
    public Xdr encode() 
    {
	Xdr xdr = super.encode();
	
	xdr.PUT(gid);
	xdr.PUT(requestMID);
	xdr.PUT(pid);
	userInfo.xdrPut(xdr);
	
	return xdr;
    }
    
    public Xdr decode()
    {
	Xdr xdr;
	
	xdr = super.decode();
	
	gid = xdr.TYPEID();
	requestMID = xdr.TYPEID();
	pid = xdr.TYPEID();
	this.userInfo = UserID.xdrGet(xdr);
	
	return xdr;
    }
    
    /**
     * Evalute a MIDIDResponseCapsule at this node.
     * <p>
     * XXX should deal with unanticipated MIDIDResponseCapsules better.
     */
    public boolean evaluate(Node node)
    {
	boolean retval;
	
	Trace.trace(Trace.ANTS_DYNLOAD, "received MIDIDResponseCapsule " +requestMID);
	
	/*
	 * Don't bother to double check that I asked for this
	 * MIDIDResponseCapsule.  If something is trying to waste
	 * our time, they can just do a better (and undetectable)
	 * job by sending fake packets with unknown MIDs.
	 */
	
	// XXX Double check that the MIDIDResponse is valid...
	// (i.e., the UserInfo...)
	
	if( node.getAddress() == getDst() )
	{
	    /* Remove the pending request for this mid */
	    node.getCache().remove(requestMID);
	    PrimordialMessenger.addMID(pid, userInfo, requestMID, gid, getSrc());
	    retval = true;
	}
	else
	{
	    retval = node.routeForNode(this, getDst());
	}
	return retval;
    }
    
    
    /**
     * Construct a blank MIDIDResponseCapsule.
     * <p>
     * This is only invoked when a MIDIDResponseCapsule capsule is
     * received at a node and the node must construct a capsule
     * before decoding the packet into it.
     */
    MIDIDResponseCapsule()
    {
    }
    
    /**
     * Construct a MIDIDResponseCapsule to send back in response
     * to a MIDIDRequestCapsule.  Only invoked by
     * MIDIDRequestCapsule.evaluate().
     *
     * @see MIDIDRequestCapsule#evaluate(Node)
     */
    MIDIDResponseCapsule(int dst,
			 TypeID unknownMID,
			 TypeID pid,
			 TypeID gid,
			 UserID userInfo)
    {
	setDst(dst);
	this.requestMID = unknownMID;
	this.pid = pid;
	this.gid = gid;
	this.userInfo = userInfo;
    }
    
    public String toString()
    {
	StringBuffer buf = new StringBuffer(100);
	
	buf.append(super.toString());
	buf.append("+MIDIDResponse[requestMID=");
	buf.append(requestMID);
	buf.append("; pid=");
	buf.append(pid);
	buf.append("; gid=");
	buf.append(gid);
	buf.append("]");
	return buf.toString();
    }
}
