/*
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 */
package ants.core;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

import edu.utah.janos.nodeos.Address;

import edu.utah.janosvm.kit.comm.commspace.CommSpaceHandle;
import edu.utah.janosvm.kit.comm.commspace.CommSpaceElement;

import ants.core.security.ReferenceMonitor;
import ants.core.security.SecurityDefaults;


/**
 * The neighbor table maintains a mapping of
 * logical active network addresses to physical Addresses.
 * The table also supports "listeners".  Each listener is notifed
 * whenever the table changes.
 * <p>
 * This class uses a somewhat crufty access control mechanism to
 * prevent access by untrusted code.  Read access is mediated by
 * performing a single reference monitor check in
 * Node.getNeighborTable. Write access is mediated by performing a
 * reference monitor check for each write access.  This optimizes for
 * the common case (reads) without explicitly separting the read and
 * write interfaces. Also, we assume that read access to the table
 * implies the right to add and remove NeighborListeners.
 * <p>
 * XXX the current implementation suffers from race conditions related
 * to multiple updates occurring frequently.  There is a basically a
 * single update buffer between the table and each listener.  If a
 * listener is not "quick enough" updates may get lost.  Additionally
 * (though less likely) updates can get partially overwritten.  
 *
 * @see NeighborListener
 * @see Node#getNeighborTable()
 * @author <a href="mailto:andrew@cs.washington.edu">Andrew Whitaker</a>
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public final class NeighborTable
{
    private static final int UPDATE_EVENT_ADD = 0x00aa;
    private static final int UPDATE_EVENT_DEL = 0x00bb;

    /**
     * The table mapping Integer ANTS addresses to Address
     * objects representing the physical link to the neighbor.
     * <p>
     * XXX should be final, but JDK1.x will fail to compile it.
     */
    private /*final*/ Hashtable neighbors;

    /**
     * The list of listeners that need to be kicked when an update
     * occurs.
     * <p>
     * XXX current implementation supports a fixed number of adds,
     * and does not support removal.
     * <p>
     * XXX should be final, but JDK1.x will fail to compile it.
     */
    private /*final*/ CommSpaceHandle listeners;
    private int listenerIndex = 0;

    /**
     * The CommSpace used to communicate updates from users of the
     * table to the root domain.
     * <p>
     * XXX should be final, but JDK1.x will fail to compile it.
     */
    private /*final*/ CommSpaceHandle updateComm;

    private static final CommSpaceElement UPDATECOMM_ADD_NEIGHBOR
	= new CommSpaceElement(0xaa);
    private static final CommSpaceElement UPDATECOMM_DEL_NEIGHBOR
	= new CommSpaceElement(0xbb);
    private static final int UPDATECOMM_OP_INDEX = 0;
    private static final int UPDATECOMM_ADDR_INDEX = 1;
    private static final int UPDATECOMM_LINK_INDEX = 2;
    private static final int UPDATECOMM_SYNC_INDEX = 3;

    /**
     * The XP (cross-process) listener.
     */
    private static class XPListener
	implements java.lang.Runnable
    {
	private final NeighborListener nl;
	private final NeighborTable table;
	private final int listenerIndex;

	public XPListener(NeighborTable table, NeighborListener nl, int index)
	{
	    this.nl = nl;
	    this.table = table;
	    this.listenerIndex = index;
	}

	public void start(Node node)
	{
	    node.systemThreadStart(this, "NeighborTable update dispatcher");
	}
	
	private int waitForUpdate()
	    throws InterruptedException
	{
	    return this.table.listenerWait(this.listenerIndex);
	}

	private Address getLinkArg()
	{
	    return this.table.getLinkArg(this.listenerIndex);
	}

	public void run()
	{
	    while(true) // XXX need a clean shutdown option?
	    {
		int addr = 0;
		Address opLink;
		Address tableLink;

		boolean got = false;
		while (!got)
		{
		    try
		    {
			addr = waitForUpdate();
			got = true;
		    }
		    catch (InterruptedException ie)
		    {
			// ignore
		    }
		}

		opLink = getLinkArg();
		
		/*
		 * Look and see if the link address is in the public
		 * table.  If it is, and it maps to the same link as
		 * we got from the commspace, then the op was an
		 * "add".  If its not there *or it is different* then
		 * the op was a delete....
		 */
		tableLink = table.getNeighbor(addr);

		if ((tableLink == null)
		    || (! tableLink.equals(opLink)))
		{
		    this.nl.neighborRemoved(addr, opLink);
		}
		else
		{
		    this.nl.neighborAdded(addr, opLink);
		}
	    }
	}

	public String toString()
	{
	    return "XPListener[idx="+listenerIndex+";nl="+nl+"]";
	}
    }

    /**
     * Only one NeighborTable is created.  It is created by the
     * PrimordialNode as it starts up.  Thus the constructor
     * is package-private.
     */
    NeighborTable()
    {
	neighbors = new Hashtable();
	listeners = CommSpaceHandle.newCommSpace(20);
	updateComm = CommSpaceHandle.newCommSpace(4);
    }

    /*PACKAGE*/ void startHandler(Node node)
    {
	node.systemThreadStart(new Runnable() {
	    public void run() {
		doManagement();
	    }}, "NeighborTable Manager");
    }

    /**
     * Return a physical address given a (virtual) active network
     * address.
     */
    public Address getNeighbor(int address)
    {
	return (Address)this.neighbors.get(new Integer(address));
    }
    
    /**
     * Return a physical address given a (virtual) active network
     * address. 
     */
    public Address getNeighbor(Integer address)
    {
	return (Address)this.neighbors.get(address);
    }
    
    /**
     * Return an enumeration of all the neighbor virtual addresses.
     * The elements of the enumeration are Integers.
     */
    public Enumeration getNeighbors()
    {
	return this.neighbors.keys();
    }
    
    public synchronized void addNeighborListener(NeighborListener nl) 
    {
	XPListener xpl;
	xpl = new XPListener(this, nl, getListenerIndex());
	xpl.start(Node.currentNode());
    }
    
    public synchronized void removeNeighborListener(NeighborListener nl)
    {
	// XXX NOT IMPLEMENTED.
    }
    
    private void kickListeners(CommSpaceElement addrCse, CommSpaceElement linkCse)
    {
	int i = 0;
	int size = listeners.size();
	
	while (i < size)
	{
	    // write link first, as server is blocked on addr index
	    listeners.write(i+1, linkCse);
	    listeners.write(i, addrCse);
	    i+=2;
	}
    }

    /**
     * Main routine for the thread dedicated to maintaining the
     * NeighborTable in the root domain's context.
     */
    private void doManagement()
    {
	final CommSpaceElement opDone = new CommSpaceElement(true);

	while (true) // XXX need clean shutdown option
	{
	    CommSpaceElement cse;
	    CommSpaceElement addrCse, linkCse;
	    CommSpaceElement op = null;
	    Integer addr;
	    Address link;

	    boolean got = false;
	    while (!got)
	    {
		try
		{
		    op = updateComm.waitToTake(UPDATECOMM_OP_INDEX);
		    got = true;
		}
		catch (InterruptedException ie)
		{
		    // ignore
		}
	    }

	    Trace.trace(Trace.ANTS_NEIGHBOR_TABLE, "Got update"
			+" op=CSE[" +op+ "]");

	    addrCse = updateComm.read(UPDATECOMM_ADDR_INDEX);
	    addr = addrCse.getInteger();
	    linkCse = updateComm.read(UPDATECOMM_LINK_INDEX);
	    //link = linkCse.getAddress();
	    link = (Address)linkCse.getCommSpaceable();
	    if (op.equals(UPDATECOMM_ADD_NEIGHBOR))
	    {
		this.neighbors.put(addr, link);
	    }
	    else if (op.equals(UPDATECOMM_DEL_NEIGHBOR))
	    {
		this.neighbors.remove(addr);
	    }
	    else
	    {
		Node.abort("WTF!?");
	    }

	    Trace.trace(Trace.ANTS_NEIGHBOR_TABLE, "kicking listeners");

	    kickListeners(addrCse, linkCse);

	    // synchronize with updater
	    Trace.trace(Trace.ANTS_NEIGHBOR_TABLE, "Synchronize with updater");
	    updateComm.write(UPDATECOMM_SYNC_INDEX, opDone);
	}
    }

    private int getListenerIndex()
    {
	Assert.assert(listenerIndex < listeners.size());

	/*
	 * Add two to the listenerIndex because +0 is the
	 * addr and +1 is the link object being operated
	 * on.
	 */
	return listenerIndex+=2;
    }

    private int listenerWait(int idx)
	throws InterruptedException
    {
	CommSpaceElement cse;
	cse = listeners.waitToTake(idx);
	return cse.getInt();
    }

    private Address getLinkArg(int idx)
    {
	CommSpaceElement cse;
	cse = listeners.read(idx+1);
	//return cse.getAddress();
	return (Address)cse.getCommSpaceable();
    }

    private void sendUpdate(CommSpaceElement op, int address, Address link)
    {
	Assert.require(op);
	Assert.require(address != 0);
	Assert.require(link);
	
	/*
	 * Send message to manager to do the op.  Write the arguments
	 * before writing the op to avoid race conditions.
	 */
	synchronized(updateComm)
	{
	    updateComm.write(UPDATECOMM_ADDR_INDEX,
			     new CommSpaceElement(address));
	    updateComm.write(UPDATECOMM_LINK_INDEX,
			     new CommSpaceElement(link));
	    updateComm.write(UPDATECOMM_OP_INDEX, op);
	    CommSpaceElement ret = null;
	    while (ret == null)
	    {
		try
		{
		    ret = updateComm.waitToTake(UPDATECOMM_SYNC_INDEX);
		}
		catch (InterruptedException ie)
		{
		    // ignore
		}
	    }
	}
	Trace.trace(Trace.ANTS_NEIGHBOR_TABLE, "Update complete "
		    +Address.toString(address)+
		    " link=" +link);
    }
    
    /**
     * add a new neighbor, and notify any listeners.  We add the new 
     * neighbor BEFORE notifying the listeners.  Requires an access
     * control check.
     */
    public void addNeighbor(int address, Address link)
    {
	Assert.require(address != 0);
	Assert.require(link);

	Trace.trace(Trace.ANTS_NEIGHBOR_TABLE, "Adding neighbor "
		    +Address.toString(address)+
		    " link=" +link);

	ReferenceMonitor.checkPermission(Node.currentPrincipal(),
					 SecurityDefaults.MODIFY_NEIGHBORS);
	
	Assert.require(link);
	
	sendUpdate(UPDATECOMM_ADD_NEIGHBOR, address, link);
    }
    
    /** 
     * remove an immediate neighbor, and notify and listeners.  We remove
     * the neighbor AFTER notifying the listeners.
     */
    public void removeNeighbor(int address)
    {
	Assert.require(address != 0);
	
	Trace.trace(Trace.ANTS_NEIGHBOR_TABLE, "Removing neighbor "
		    +Address.toString(address));

	ReferenceMonitor.checkPermission(Node.currentPrincipal(),
					 SecurityDefaults.MODIFY_NEIGHBORS);
	
	Integer key = new Integer(address);
	Address link = (Address)this.neighbors.get(key);

	Node.currentNode().remOutChannel(address);
	if (link != null)
	{
	    sendUpdate(UPDATECOMM_DEL_NEIGHBOR, address, link);
	}
    }
}
