/* 
 * Node.java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * Copyright (c) 2000, 2001 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package ants.core;

import edu.utah.janos.nodeos.Address;
import edu.utah.janos.nodeos.AddressFormatException;
import edu.utah.janos.nodeos.BufferHandle;
import edu.utah.janos.nodeos.NoSuchNodeAddress;
import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.InChannel;
import edu.utah.janos.nodeos.OutChannel;
import edu.utah.janos.nodeos.Credentials;
import edu.utah.janos.nodeos.Thread;
import edu.utah.janos.nodeos.ThreadPoolSpec;
import edu.utah.janos.nodeos.MemSpec;
import edu.utah.janos.nodeos.ProtocolSpecFormatException;
import edu.utah.janos.nodeos.PacketAddressFormatException;

import edu.utah.janos.io.Log;
import edu.utah.janos.io.LogListener;

import edu.utah.janosvm.kit.comm.commspace.CommSpaceHandle;
import edu.utah.janosvm.kit.comm.commspace.CommSpaceElement;
import edu.utah.janosvm.kit.comm.commspace.CommQueuePusher;

import java.io.IOException;

import java.util.Hashtable;
import java.util.Enumeration;

import ants.core.security.Principal;
import ants.core.security.SecurityDefaults;
import ants.core.security.ReferenceMonitor;

/**
 * The per-domain Node interface.
 * <p>
 * An instance of Node is created in the context of each Application
 * and protocol.
 *
 * @see PrimordialNode
 * @see AppBooster
 * @see ProtocolBooster
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class Node
{
    /**
     * Mapping from Integer() active address to out channel for each
     * neighbor this node can access.
     */
    private final Hashtable outChannels = new Hashtable(12, 0.5f);
    
    private final Log logger;
   
    /**
     * The security principal associated with this domain
     */
    private Principal principal;
    
    /**
     * Table for domain-private interned values.
     */
    private final Hashtable internTable = new Hashtable();
    
    /*package*/ final UnknownMidPackets unknownMidPackets;
    
    /**
     * Maps groupIDs to CodeGroups.
     */
    private final Hashtable GIDtoCodeGroup = new Hashtable();
    
    /**
     * Maps MIDs to CapsuleClasses.
     */
    private final Hashtable MIDtoCapsuleClass = new Hashtable();
    
    /**
     * Maps name to CapsuleClass.  This table is used to supply
     * a PID or MID given the name of a CapsuleClass.
     */
    private final Hashtable nameToCapsuleClass = new Hashtable(20,0.6f);
    
    /**
     * A "list" of all the applications running on this Node (i.e.,
     * in this Domain.
     * <p>
     * Optimized for the current case of exactly one app per node...
     * <p>
     * Will be null in protocol-based domains.
     */
    private Application app = null;
    
    private NodeCache cache;

    private BufferHandle bufferHandle = BufferHandle.create();

    private Domain domain;
    
    /**
     * Construct a node.  A node is only constructed by the boot
     * routine in PrimordialNode, AppBooster or ProtocolBooster.
     * <p>
     * Note that the Principal provided can be overriden later via
     * the setPrincipal() call.  New nodes generally start with a
     * reduced-admin principal, which is dropped to a remote-user
     * or local-user principal when the node is standing on its
     * feet (and can figure out who it is).
     *
     * @param domain the new domain this Node is associated with.  Must not be null.
     * @param address the ANTS virtual address of this node.  XXX redundant.
     * @param unknownMidPackets the storage that holds packets this new Node should process.
     * @param principal the initial principal this node should execute as.
     */
    Node(Domain domain,
	 UnknownMidPackets unknownMidPackets,
	 Principal principal)
    {
	Assert.require(domain);
	Assert.require(principal);
	
	// Put node object where it can be looked up.
	domain.setObject(this);

	this.domain = domain;
	this.principal = principal;
	this.unknownMidPackets = unknownMidPackets;
	// XXX perhaps the cache should be started on demand? (in getCache())
	this.cache = new NodeCache(this, 15);
	this.cache.collect();
	// this.logger = new Log(domain.getName());
	this.logger = new Log(null);
	this.logger.addListener(new PrimordialNode.StdoutListener());
    }
    
    /** 
     * Return the security principal associated with this node
     *
     * @see ants.core.security.Principal
     */
    public Principal getPrincipal()
    {
	Assert.assert(principal != null);
	return principal;
    }
    
    /**
     * Set the principal associated with this node.  This
     * should only be invoked once per Node.
     */
    void setPrincipal(Principal p)
    {
	Assert.require(p);
	
	/*
	 * As we use it, the principal is only ever changed
	 * from the bootstrap user to something else.
	 */
	Assert.assert(p != SecurityDefaults.BOOTSTRAP_USER);
	Assert.assert((this.principal == SecurityDefaults.BOOTSTRAP_USER) ||
		      (this.principal == SecurityDefaults.REMOTE_USERS));
	
	this.principal = p;
    }
    
    
    /**
     * Return the logical 32-bit IP-like "active" address for this
     * Node.  May return 0, which means the address for the node has
     * not yet been assigned.  That happens with DANTE or other
     * dynamic address specification protocols.
     *
     * @return the 32-bit IP-like "active" address of this Node
     */
    public int getAddress()
    {
	return PrimordialNode.getAddress();
    }
    
    /**
     * Set the logical 32-bit address for this (and every) node
     * This method assumes synchronization is provided elsewhere
     * if this is necessary.
     */
    public void setAddress (int a)
    {
	ReferenceMonitor.checkPermission(currentPrincipal(),
					 SecurityDefaults.SET_ADDRESS);
	PrimordialNode.setAddress(a);
    }
    
    
    /**
     * Get the protocol spec that remote nodes can use to communicate
     * with this node.  This protocol spec will be the spec for the
     * proxy, if this node is being proxied.
     *
     * @see PrimordialNode#getVisibleProtocolSpec()
     */
    public String getVisibleProtocolSpec()
    {
	
	ReferenceMonitor.checkPermission(currentPrincipal(),
					 SecurityDefaults.READ_PHYSICAL_PROTOCOL_SPEC);
	
	return PrimordialNode.getVisibleProtocolSpec();
    }
    
    /**
     * Get the address spec that remote nodes can use to communicate
     * with this node.  This address spec will be the spec for the
     * proxy, if this node is being proxied.
     *
     * @see PrimordialNode#getVisibleProtocolSpec()
     */
    public String getVisibleAddressSpec()
    {
	
	ReferenceMonitor.checkPermission(currentPrincipal(),
					 SecurityDefaults.READ_PHYSICAL_ADDRESS_SPEC);
	
	return PrimordialNode.getVisibleAddressSpec();
    }
    
    /**
     * Return the Node interface associated with the current domain.
     * <p>
     * XXX This method is a hack and would go away if
     * one were to re-write the basic ANTS interfaces.  The
     * "current node" should be an explicit paramter to those
     * functions that need it.
     */
    public static Node currentNode()
    {
	Domain f;
	Object n;
	
	f = Domain.currentDomain();
	Assert.assert(f != null, "currentDomain() non-null");
	n = f.getObject();
	Assert.assert(n != null, "currentDomain() has non-null object");
	Assert.assert(n instanceof Node, "currentDomain() object is a Node object");
	
	return (Node)n;
    }
    
    public static final Principal currentPrincipal() 
    {
	return currentNode().getPrincipal();
    }

    private OutChannel getOutChannel(int logicalAddress)
    {
	return this.getOutChannel(logicalAddress,
				  PrimordialNode.
				  nextPhysicalToNode(logicalAddress));
    }
    
    private OutChannel getOutChannel(int logicalAddress,
				     Address physicalAddress)
    {
	if( Trace.ENABLED )
	{
	    Trace.trace(Trace.ANTS_STARTUP, "Node.getOutChannel maps "
			+Address.toString(logicalAddress)+
			" to " +physicalAddress);
	}
	
	Integer key = new Integer(logicalAddress);
	OutChannel retval = null;
	NeighborOutChannel noc;

	if( ((noc = (NeighborOutChannel)this.outChannels.get(key)) == null) ||
	    !noc.valid(physicalAddress) )
	{
	    try
	    {
		if( noc == null )
		    noc = new NeighborOutChannel(physicalAddress);
		else
		    noc.setNeighbor(physicalAddress);
	    }
	    catch (java.io.IOException ioe)
	    {
		Node.abort(ioe, "OutChannel.create() threw: ");
	    }
	    this.outChannels.put(key, noc);
	}
	retval = noc.getOutChannel();
	return retval;
    }

    void remOutChannel(int logicalAddress)
    {
	NeighborOutChannel noc;

	noc = (NeighborOutChannel)outChannels.
	    remove(new Integer(logicalAddress));
	noc.destroy();
    }
    
    /**
     * Just calls {@link #register(Protocol,Principal) register(Protocol,Principal)}
     * with a default Principal (SecurityDefaults.REMOTE_USERS).
     *
     * @see SecurityDefaults.REMOTE_USERS
     * @see #register(Protocol,Principal)
     */
    public void register(Protocol p)
    {
	register(p, SecurityDefaults.REMOTE_USERS);
    }

    /**
     * Called by applications that want to instantiate a
     * locally originiating protocol.  No security check here
     * because we assume that dynamically loaded protocol code
     * does not have access to the Protocol class.
     */
    public void register(Protocol protocol, Principal principal)
    {
	Trace.trace(Trace.ANTS_DYNLOAD, "Installing protocol " +protocol+
		    "(with princial="+principal+")");
	try
	{
	    // Install the Protocol in its own domain.
	    protocol.activate(this, true, principal);
	}
	catch (Exception e)
	{
	    Node.abort(e, "Protocol registration blew up:");
	}
    }
    
    /**
     * Deliver the given capsule to the Application on the
     * given port number.
     * <p>
     *
     * @exception NoSuchApplicationError if the port doesn't have an app listening on it.
     */
    public boolean deliverToApp(Capsule cap, int port)
    {
	CommQueuePusher app = PrimordialNode.getAppQueue(port);
	
	if (app == null)
	    throw new NoSuchApplicationError("There is no application registered on port " +port);
	
	CommSpaceElement msg[] = { null, null };
	
	cap.enqueueCapsule(msg, 0);
	app.enqueue(msg);
	return true;
    }
    
    /**
     * Inject a given capsule into the network via the local node.  
     * This method is only called by Application#send().
     * Note that the capsule's evaluate() is responsible for routing...
     */
    public boolean send(Capsule cap)
    {
	ReferenceMonitor.checkPermission
	    (currentPrincipal(),SecurityDefaults.LOCAL_SEND);
	
	return cap.evaluate(this);
    }
    
    /**
     * @see #threadStart(java.lang.Runnable,String)
     */
    public void threadStart(java.lang.Runnable r)
    {
	Assert.require(r);
	
	threadStart(r, "Another Underappreciated Thread");
    }
    
    /**
     * Start a new thread executing the given Runnable.
     * The caller must have the START_THREAD permission.
     */
    public void threadStart(java.lang.Runnable r, String threadName)
    {
	Assert.require(r);
	
	ReferenceMonitor.checkPermission
	    (currentPrincipal(),SecurityDefaults.START_THREAD);

	systemThreadStart(r, threadName);
    }

    /**
     * Start a thread.  No security check is done, so this should
     * only be used for "system" threads.
     */
    /*PACKAGE*/ void systemThreadStart(java.lang.Runnable r, String threadName)
    {
	Assert.require(r);
	
	Domain.currentDomain().threadStart(r, threadName);
    }
    
    /**
     * Route a capsule for a particular destination.
     * Note that there is no next hop to the current node, and it is
     * therefore impossible to route to ourself.  Since capsules are
     * evaluated at the local node before being sent, it seems like a bad
     * idea to allow them to route to the same node -- this would likely
     * produce an infinite loop.
     *
     */
    public synchronized boolean routeForNode(Capsule cap, int destination)
    {
	Xdr xdr;
	
	if (Trace.ENABLED)
	    Trace.trace(Trace.ANTS_OUTGOING, "Sending capsule " +cap+ " to " +Address.toString(destination));
	
	Assert.assert(destination != 0);
	
	// Encode the capsule.
	xdr = cap.encodeCapsuleForSend(this);
	
	int target;
	
	if( destination == PrimordialNode.getAddress() )
	{
	    Trace.trace(Trace.ANTS_OUTGOING, "  Warning: Capsule is being routed to this node");
	    return false;
	}
	else
	{
	    // look up nearest neighbor on the way to destination
	    target = PrimordialNode.nextHopToNode(destination);
	}
	
	if (target == 0) // XXX what to do?
	{
	    Trace.trace(Trace.ANTS_OUTGOING, "  Capsule send failed.  No hop to " +Address.toString(destination));
	    return false;
	}
	
	if(Trace.ENABLED)
	    Trace.trace(Trace.ANTS_OUTGOING, "  Capsule being forwarded to " +Address.toString(target));
	
	
	// Push out the appropriate out-channel
	Object oChanObj = this.getOutChannel(target);
	
	Assert.assert(oChanObj != null);
	
	// XXX fix when XDR is fixed!
	int xdrLen = xdr.length();

	this.bufferHandle.setLength(xdrLen);
	this.bufferHandle.setBytes(0, xdr.getBytes(), 0, xdrLen);
	
	if (Trace.ENABLED)
	    Trace.trace(Trace.ANTS_OUTGOING, "  Pushing buf " + this.bufferHandle.toString());
	((OutChannel)oChanObj).send(this.bufferHandle);
	
	PrimordialNode.getStats().sentPacket();
	
	return true;
    }
    
    /**
     * Send a capsule to an adjacent node.  This method does not
     * rely on the routing protocol, thus its quite useful for
     * BOOTP-like protocols that need to send a packet to discover
     * the Node's address.
     * <p>
     * Note that an association between the neighbor's logical
     * address and a physical address must have been established and
     * installed in the PrimoridalNode's neighborTable.
     * <p>
     * Note that because the capsules sent this way are used for
     * bootstrapping, they have no source address.  This routine
     * inserts "255.255.255.255" as the source addr, and fixes the
     * TTL at 3.  Because of this, this routine should only be
     * used by privilged bootstrap protocols.
     * 
     * @param cap the capsule to send.  Will be encoded and launched
     * just like any other capsule.
     * @param neighbor the logical address of the neighbor node.
     * @see PrimordialNode#neighborTable
     * @see PrimordialNode#loadRouteTableImpl(int,boolean,String)
     */
    public boolean sendToNeighbor(Capsule cap, int neighbor) {
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	ReferenceMonitor.checkPermission
	    (currentPrincipal(),SecurityDefaults.RAW_SEND);
	
	Xdr xdr;
	
	if (Trace.ENABLED)
	    Trace.trace(Trace.ANTS_OUTGOING, "Sending capsule " +cap+ " to neighbor " +Address.toString(neighbor));
	
	Assert.assert(neighbor != 0);
	
	/*
	 * Put a fake src and small ttl in the packet if it
	 * coming from the client
	 */
	if (cap.getSrc() == 0) 
	{
	    cap.prime(-1, 3);
	    cap.setDst(neighbor);
	}
	
	// Encode the capsule.
	xdr = cap.encodeCapsuleForSend(this);
	
	Address target = null;
	
	if( neighbor == this.getAddress() )
	{
	    Trace.trace(Trace.ANTS_OUTGOING, "  Warning: Capsule is being routed to this node");
	    return false;
	}
	else
	{
	    // look up the neighbor's physical address
	    target = PrimordialNode.nextPhysicalToNeighbor(neighbor);
	}
	
	if (target == null) // XXX what to do?
	{
	    Trace.trace(Trace.ANTS_OUTGOING, "  Capsule send failed.  No hop to " +Address.toString(neighbor));
	    return false;
	}
	
	if (Trace.ENABLED)
	    Trace.trace(Trace.ANTS_OUTGOING, "  Capsule being forwarded to " +Address.toString(neighbor));
	
	
	// push out the appropriate out-channel
	Object oChanObj = this.getOutChannel(neighbor);

	Assert.assert(oChanObj != null);
	
	// XXX fix when XDR is fixed!
	BufferHandle bh = BufferHandle.create(0, xdr.length());
	bh.setBytes(0, xdr.getBytes(), 0, xdr.length());
	
	if (Trace.ENABLED)
	    Trace.trace(Trace.ANTS_OUTGOING, "  Pushing buf " +bh.toString());
	((OutChannel)oChanObj).send(bh);
	
	return true;
    }
    
    
    /**
     * Publish the given CapsuleClass in both the name -> cc map and the
     * MID -> cc map.
     *
     * @param cc the CapsuleClass to publish.  Must have a name and methodID defined.
     * @exception DuplicateCapsuleClassException if a CC is already registered under the cc's name.
     */
    void publishCapsuleClass(CapsuleClass cc)
	throws DuplicateCapsuleClassException
    {
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	Object r;
	Assert.assert(cc);
	Assert.assert(cc.name());
	Assert.assert(cc.methodID);
	
	Trace.trace(Trace.ANTS_DYNLOAD, "publishing " + cc);
	
	synchronized(nameToCapsuleClass)
	{
	    // Publish in the name -> cc map
	    r = nameToCapsuleClass.put(cc.name(), cc);
	    if (r != null)
	    {
		Object o;
				// put the original element back
		o = nameToCapsuleClass.put(cc.name(), r);
		Assert.assert(o == cc);
		throw new DuplicateCapsuleClassException("Tried to re-register " + cc + ".");
	    }
	}
	
	// Publish in the mid -> cc map
	r = MIDtoCapsuleClass.put(cc.methodID, cc);
	Assert.assert(r == null);
    }
    
    void publishCodeGroup(CodeGroup group)
    {
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	Object rc;
	
	Trace.trace(Trace.ANTS_DYNLOAD, "publishing " + group);
	rc = GIDtoCodeGroup.put(group.groupID, group);
	Assert.assert(rc == null);  // no duplicates, please.
    }
    
    TypeID findMID(String name)
    {
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	CapsuleClass cc;
	cc = (CapsuleClass)nameToCapsuleClass.get(name);
	if (cc == null)
	    Node.abort("No CapsuleClass registered under " +name+ " (in findMID()).");
	
	Assert.assert(cc.methodID);
	return cc.methodID;
    }
    
    TypeID findPID(String name)
    {
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	CapsuleClass cc;
	cc = (CapsuleClass)nameToCapsuleClass.get(name);
	if (cc == null)
	    Node.abort("No CapsuleClass registered under " +name+ " (in findPID()).");
	
	Assert.assert(cc.protocolID());
	return cc.protocolID();
    }
    
    /**
     * Find a CapsuleClass for the given MID.  CapsuleClasses are
     * registered by <code>publishCapsuleClass</code>.
     *
     * @see #publishCapsuleClass
     */
    CapsuleClass lookupCapsuleClass(TypeID mid)
    {
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	Trace.trace(Trace.ANTS_DYNLOAD, "cc for mid " +mid+ " -> " +MIDtoCapsuleClass.get(mid));
	
	return (CapsuleClass) MIDtoCapsuleClass.get(mid);
    }
    
    CodeGroup lookupCodeGroup(TypeID gid)
    {
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	Trace.trace(Trace.ANTS_DYNLOAD, "CodeGroup for gid " +gid+ " -> " +GIDtoCodeGroup.get(gid));
	return (CodeGroup) GIDtoCodeGroup.get(gid);
    }
    
    void attachApplication(Application app)
    {
	this.app = app;
    }
    
    public NodeCache getCache()
    {
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	return this.cache;
    }
    
    /**
     * Have the calling thread sleep for at least the given
     * number of milliseconds.
     *
     * @param millis how long to sleep for, in milliseconds.
     */
    public void sleep(long millis)
	throws InterruptedException
    {
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	edu.utah.janos.nodeos.Thread.sleep(millis, 0);
    }
    
    
    /**
     * Shutdown this Node.  Just kills the domain, and not
     * the entire virtual machine.
     * Shuts down the cache, and de-register's this
     * domain's application if it has one.  To shutdown
     * the entire VM, use the non-public PrimordialNode.shutdown().
     * <p>
     * Interface for shutting down the entire node
     * isn't (yet) visible.
     *
     * @see PrimordialNode#shutdown()
     */
    public void shutdown()
    {
	
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	this.cache.stopCollecting();
	// If an app is registered by this Node, remove it
	if (this.app != null)
	{
	    PrimordialMessenger.detachApplication(app.getPort());
	    
	    // If a one-trick show, shut down the whole VM
	    if (PrimordialNode.oneTrick())
		PrimordialNode.shutdown();
	}
	
	f.destroy();
    }
    
    /**
     * Return the string name of the host machine.
     */
    public String hostname()
    {
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	return edu.utah.janos.nodeos.Node.hostname();
    }
    
    /**
     * Return the current time in milliseconds.
     */
    public long time()
    {
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject(), "node method being called in correct domain");
	
	return edu.utah.janos.nodeos.Node.time();
    }
    
    /**
     * Log a message to this Node's "console".  The
     * messages are sent to the logger installed
     * on this node.  The level parameter indicates
     * the serverity of the condition the message relays.
     * Use constants from the
     * <code>edu.utah.janos.io.Log</code> class for this parameter
     *
     * @param level the severity/priority of the log message
     * @param msg the string message
     * @see #logger
     * @see edu.utah.janos.io.Log
     */
    public void log(int level, String msg)
    {
	
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	ReferenceMonitor.checkPermission
	    (currentPrincipal(),SecurityDefaults.LOG);	       
	
	this.logger.logMsg(level, msg);
    }
    
    /**
     * Log an exception.  Level parameter is same as for logging
     * a string, see log(int, String)
     * <p>
     * XXX redundant with logStackTrace(int,Throwable)
     *
     * @see #log(int,String)
     */
    public void log(int level, Throwable exception)
    {
	
	ReferenceMonitor.checkPermission
	    (currentPrincipal(),SecurityDefaults.LOG);
	
	logStackTrace(level, exception);
    }
    
    /**
     * Log an exception.  Same as log(int, Throwable)
     * <p>
     * XXX redundant with log(int,Throwable)
     *
     * @see #log(int,Throwable)
     */
    public void logStackTrace(int level, Throwable exception)
    {
	
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	ReferenceMonitor.checkPermission
	    (currentPrincipal(),SecurityDefaults.LOG);
	
	
	this.logger.logThrowable(level, exception);
    }
    
    public RouteTable getRouteTable()
    {
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	/*
	 * Access to the RouteTable object is implicitly read/write,
	 * so we check write permission here.  Doing the check here
	 * makes revocation of write access difficult.
	 */
	ReferenceMonitor.checkPermission
	    (currentPrincipal(),SecurityDefaults.MODIFY_ROUTES);
	
	return PrimordialNode.getRouteTable();
    }
    
    public NeighborTable getNeighborTable() {
	Domain f = Domain.currentDomain();
	Assert.assert(this == f.getObject());
	
	/*
	 * Only check for READ permission at this time.  Each write to
	 * the RouteTable has a separate permission check.  Doing the
	 * check here makes revocation difficult.
	 */
	ReferenceMonitor.checkPermission
	    (currentPrincipal(),SecurityDefaults.READ_NEIGHBORS);   
	
	return PrimordialNode.getNeighborTable();
    }
    
    Hashtable getInternTable()
    {
	return this.internTable;
    }
    
    /**
     * @param index The statistic to retrieve, the value should be one of
     * the NodeStat constants.
     * @return The value of the statistic.
     */
    public double getStat(int index)
    {
	return PrimordialNode.getStats().getStat(index);
    }

    public void arraycopy(Object src, int src_position,
			  Object dst, int dst_position, int length)
    {
	System.arraycopy(src, src_position, dst, dst_position, length);
    }

    public String toString()
    {
	return "Node[a=" + Address.toString(this.getAddress()) + "]";
    }
    
    /**
     * Abort execution of the entire VM.  This routine prints
     * the given message to the PrimordialNode's logger with
     * the priority Log.EMERG, and then causes the VM to exit.
     * <p>
     * XXX this should move to PrimordialNode.
     */
    /*package*/ static void abort(String msg)
    {
	PrimordialNode.message(Log.EMERG, "ABORT:");
	PrimordialNode.message(Log.EMERG, msg);
	PrimordialNode.dumpStackTrace(Log.EMERG, new Throwable("just a backtrace"));
	System.exit(11);
    }
    
    /**
     * Abort execution of the entire VM.  This routine prints
     * the given message to the PrimordialNode's logger with
     * the priority Log.EMERG, followed by the the given
     * exception, and then causes the VM to exit.
     * <p>
     * XXX this should move to PrimordialNode.
     */
    /*package*/ static void abort(Throwable th, String msg)
    {
	PrimordialNode.message(Log.EMERG, "ABORT:");
	PrimordialNode.message(Log.EMERG, msg);
	PrimordialNode.dumpStackTrace(Log.EMERG, th);
	System.exit(11);
    }
}
