/* 
 * NodeConfigure.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 * 
 */
package ants.core;

import ants.util.KeyArgs;
import ants.util.InvalidKeyArgsException;
import ants.util.LineListTokenizer;
import ants.util.LineListException;

import edu.utah.janos.nodeos.Address;
import edu.utah.janos.nodeos.AddressFormatException;
import edu.utah.janos.nodeos.NoSuchNodeAddress;
import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.DomainRun;
import edu.utah.janos.nodeos.DomainStartException;

import edu.utah.janos.io.Log;

import edu.utah.janosvm.kit.comm.commspace.CommSpaceHandle;

import java.io.FileNotFoundException; // XXX
import java.io.IOException; // XXX

import java.util.Vector; // XXX
import java.util.StringTokenizer; // XXX

/**
 * Performs one-time configuration of a Node to get it on its feet.  This is
 * one of the ways to start an ANTS node.
 * <p>
 * An alternative to ConfigurationManager.
 * <p>
 * Usage:
 * <code>NodeConfigure <var>ActiveAddress</var> <var>PhysicalAddress</var> [<var>options</var>]</code>
 * <menu>
 *   <li> <b>ActiveAddress</b>: the active address of the Node.  Must be an IP-like dotted-quad.
 *   <li> <b>PhysicalAddress</b>: the physical host address of the Node.  Must be a <var>host:port</var> pair.
 * </menu>
 * Options:
 * <menu>
 *   <li> <b>-consoleport <var>int</var></b>: Start a telnet console on the given port.
 *   <li> <b>-routes <var>file</var></b>: Static routing table is in <var>file</var>.
 *   <li> <b>-updateRoutes</b>: If dynamic routing updates should be pushed out.
 *   <li> <b>-application <var>appfile</var></b>: Read <var>appfile</var> to configure & start an application.
 *   <li> <b>-simpleapp <var>string</var></b>: Use the string as an app and its arguments.
 *   <li> <b>-extension <var>extfile</var></b>: Read <var>extfile</var> to configure and install an extension.
 *   <li> <b>-msgprefix <var>string</var></b>: Prefix all messages with <var>string</var>.
 *   <li> <b>-trace <var>flags</var></b>: Set the node trace flags to <var>flags</var>.
 *   <li> <b>-policy <var>file</var></b>: XXX this needs to be written.
 *   <li> <b>-principal <var>name</var></b>: XXX this needs to be written.
 *   <li> <b>-proxyaddr <var>addr</var></b>: XXX this needs to be written.
 * </menu>
 * The files <var>appfile</var> and <var>extfile</var> are simple files that are unceremoniously
 * pushed into KeyArgs.  They should define the application.
 * <p>
 * XXX The physical address should be specified in a NodeOS-compliant syntax.<br>
 * 
 *
 * @see ConfigurationManager
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class NodeConfigure
    extends DomainRun
{
    /**
     * XXX A hack.  Move into PrimordialNode so that ConfigurationManager and NodeConfigure
     * will be mutually exclusive.
     */
    private static boolean beenHere = false;
    
    /**
     * Configure this Node.  Prints a usage message if
     * insufficient arguments are given.
     * <p>
     * The domain that starts this ball rolling will become the basic
     * domain in the system.
     * <p>
     * XXX domain argument isn't used anymore?
     *
     * @param args the node configuration.  See the class documentation.
     */
    protected void run(Domain domain,
		       String[] args,
		       CommSpaceHandle commSpaceHandle)
    {
	Assert.require(domain == Domain.currentDomain());
	
	if (beenHere)
	{
	    // XXX throw new AlreadyInitializedException() ??
	    Node.abort("Already initialized.  Don't be doing this.");
	}
	beenHere = true;
	
	Domain.currentDomain().setName("Primordial ANTS Domain");
	
	if ((args == null) || (args.length < 2))
	{
	    printUsage();
	    System.exit(1);
	}
	
	/*
	 * Parse the Node configuration, find what apps to start (if any)
	 * what extensions to install (if any), and then start the 
	 * PrimordialNode.
	 */
	try
	{
	    parseCommandLine(args);
	}
	catch(Exception e)
	{
	    Node.abort(e, "Bad configuration (node " +args[0]+ " " +args[1]+ ").");
	}
    }
    
    private static void printOpt(String opt, String desc)
    {
	PrimordialNode.message(Log.EMERG,
			       "    " +opt+ ":\t" + desc);
    }
    
    private static void printUsage()
    {
	PrimordialNode.message(Log.EMERG, "usage: " 
			       +NodeConfigure.class.getName()+
			       " <activeaddr> <physaddr> [options].");
	printOpt("-consoleport <int>", "Start a telnet-able console on the given port.");
	printOpt("-routes <file>", "Static routing table is in given file.");
	printOpt("-updateRoutes", "Turn on dymanic routing updates.");
	printOpt("-application <appfile>", "Read given file to configure & start an application.");
	printOpt("-simpleapp <string>", "Start a simple application, string is tokenized for arguments.");
	printOpt("-extension <extfile>", "Read given file to configure and install an extension [NOT IMPL].");
	printOpt("-msgprefix <string>", "Prefix all console messages with the given string.");
    }
    
    /**
     * Parse the command line arguments to NodeConfigure.  Most
     * arguments are passed into PrimordialNode in some fashion.
     * However, -application, -simpleapp and -extension arguments have their
     * arguments parsed and put into the <code>apps</code> and
     * <code>exts</code> return parameters.
     */
    private void parseCommandLine(String[] args)
	throws FileNotFoundException,
	IOException,
	InvalidKeyArgsException,
	AddressFormatException,
	NoSuchNodeAddress,
	InterruptedException,
	LineListException,
	ClassNotFoundException,
	DomainStartException
    {
	Vector apps = new Vector();
	Vector exts = new Vector();
	String physAddr;
	String activeAddr;
	String host = "";
	String transport = "udp";
	int port = -1;
	KeyArgs primNodeArgs;
	int i;
	
	Assert.assert(args.length >= 2);
	Assert.assert(apps);
	Assert.assert(exts);
	
	// Grab two fixed args
	activeAddr = args[0];
	physAddr = args[1];
	
	primNodeArgs = new KeyArgs();
	
	i = 2;
	while (i < args.length)
	{
	    if (args[i].equals("-application"))
	    {
		String[] appargs;
		String fname;
		i++;
		Assert.assert(i < args.length); // XXX throw error
		
		fname = args[i];
		
		try
		{
		    appargs = grokArgFile(fname);
		    apps.addElement(new AppConfig(appargs[0], appargs, 1));
		}
		catch (Exception e)
		{
		    Node.abort(e, "groking -application arg " +fname);
		}
	    }
	    else if (args[i].equals("-simpleapp"))
	    {
		String[] appargs;
		AppConfig appCfg;
		
		i++;
		Assert.assert(i < args.length); // XXX throw an error
		
		StringTokenizer stok = new StringTokenizer(args[i]);
		appargs = new String[stok.countTokens()];
		int x = 0;
		while(stok.hasMoreTokens())
		    appargs[x++] = stok.nextToken();
		try
		{
		    appCfg = new AppConfig(appargs[0], appargs, 1);
		    apps.addElement(appCfg);
		}
		catch (Exception e)
		{
		    Node.abort(e, "groking -simpleapp arg.");
		}
	    }
	    else if (args[i].equals("-extension"))
	    {
		String[] extargs;
		String fname;
		
		i++;
		Assert.assert(i < args.length); // XXX
		fname = args[i];
		
		try
		{
		    extargs = grokArgFile(fname);
		    exts.addElement(extargs); // XXX XXX XXX
		}
		catch (Exception e)
		{
		    Node.abort(e, "groking -extension arg " +fname);
		}
	    }
	    else if (args[i].equals("-updateRoutes"))
	    {
		primNodeArgs.set(args[i], null);
	    }
	    else
	    {
		String arg;
		String param;
		arg = args[i];
		i++;
		Assert.assert(i < args.length); // XXX
		param = args[i];
		
		primNodeArgs.set(arg, param);
	    }
	    i++;
	}
	
	int sep = physAddr.indexOf(':');
	if (sep > 0)
	{
	    int tsep = physAddr.indexOf(':', sep + 1);
	    try
	    {
		host = physAddr.substring(0,sep);
		port = Integer.parseInt(physAddr.substring(sep+1,
							   (tsep > 0) ?
							   tsep :
							   physAddr.length()));
		if( tsep > 0 )
		    transport = physAddr.
			substring(tsep+1,
				  physAddr.length());
				// above might throw NumberFormatException
	    }
	    catch (Exception e)
	    {
		Node.abort(e, "Problem parsing physical address: ");
	    }
	}
	else
	{
	    Node.abort("No port (or no host) provided in physical address '" +physAddr+ "'");
	}
	
	Trace.trace(Trace.ANTS_STARTUP, "Node Configuration:");
	Trace.trace(Trace.ANTS_STARTUP, "  ANTS address    : " +activeAddr);
	Trace.trace(Trace.ANTS_STARTUP, "  physical address: " +physAddr);
	Trace.trace(Trace.ANTS_STARTUP, "  node args: " +primNodeArgs);
	
	if (apps.size() > 0)
	{
	    for(i = 0; i < apps.size(); i++)
	    {
		Trace.trace(Trace.ANTS_STARTUP, "  app: " +apps.elementAt(i));
	    }
	}
	else
	{
	    Trace.trace(Trace.ANTS_STARTUP, "  [no applications]");
	}
	
	if (exts.size() > 0)
	{
	    for(i = 0; i < exts.size(); i++)
	    {
		Trace.trace(Trace.ANTS_STARTUP, "  extension: " +exts.elementAt(i));
	    }
	}
	else
	{
	    Trace.trace(Trace.ANTS_STARTUP, "  [no extensions]");
	}
	
	
	PrimordialNode.setAddress(Address.fromString(activeAddr));
	PrimordialNode.setPhysicalAddress("if/ipv4/" +transport,
					  "*:" +port+ ":" +host);
	PrimordialNode.setArgs(primNodeArgs);
	
	PrimordialNode.start(apps, exts);
	
    }
    
    /**
     * Grok an arg file (for a weak definition of grok).  Just
     * turns the contents of the file into an array of Strings,
     * one for each token.
     */
    private static String[] grokArgFile(String filename)
	throws FileNotFoundException,
	IOException,
	LineListException
    {
	LineListTokenizer lt;
	String[] line;
	String[] buf = new String[64];
	String[] ret;
	int start;
	int end;
	
	start = 0;
	end = buf.length;
	
	lt = new LineListTokenizer(filename);
	do
	{
	    line = lt.list();
	    
	    if (line != null)
	    {
		Assert.assert((start + line.length) < end);
		
				// Copy line array in
		System.arraycopy(line, 0,
				 buf, start, line.length);
		start += line.length;
	    }
	}
	while (line != null);
	
	// Dont return all 64 arguments
	ret=new String[start];
	System.arraycopy(buf,0,ret,0,start);
	return ret;
    }
    
    /**
     * This doesn't work.  Its just here to prevent folks from getting
     * really confused.
     *
     * @see #run(Domain,String[],CommSpaceHandle)
     */
    public static void main(String[] args)
	throws Exception
    {
	PrimordialNode.message(Log.EMERG, "NodeConfigure must be started as a Janos Domain.");
	System.exit(11);
    }
}
