/*
 * NodeStat.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core;

import edu.utah.janos.nodeos.Domain;

/**
 * The NodeStat class is used to store statistics for an entire node.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class NodeStat
{
    /*
     * The various statistics we currently track.
     */
    public static final int STARTUP_TIME		= 0;
    public static final int PACKETS_RECEIVED	= 1;
    public static final int PACKETS_SENT		= 2;
    public static final int AVG_PACKETS_PER_SECOND	= 3;
    public static final int MAX_PACKETS_PER_SECOND	= 4;
    public static final int STAT_MAX		= 5;
    
    /**
     * The statistics we're tracking.
     */
    private double stats[] = new double[STAT_MAX];
    
    /**
     * @return The current time in milliseconds
     */
    private long time()
    {
	return edu.utah.janos.nodeos.Node.time();
    }
    
    /**
     * This function recomputes any statistics that cover a certain
     * window of time.
     *
     * @param oldStats The values of the statistics at the start of the
     * sample window.
     * @param newStats The values of the statistics at the end of the
     * sample window.
     * @param currTime The current time value.
     * @param diffTime The size of the sample window.
     */
    private void recomputeStats(double oldStats[], double newStats[],
				long currTime, long diffTime)
    {
	double diff, tmp, timeMultiplier;
	long elapsedSeconds;
	
	timeMultiplier = ((double)1000.0) / ((double)diffTime);
	elapsedSeconds =
	    (currTime - (long)this.stats[STARTUP_TIME]) /
	    1000;
	this.stats[AVG_PACKETS_PER_SECOND] =
	    newStats[PACKETS_RECEIVED] / elapsedSeconds;
	diff = newStats[PACKETS_RECEIVED] - oldStats[PACKETS_RECEIVED];
	if( diff != 0.0 )
	{
	    tmp = diff * timeMultiplier;
	    if( tmp > newStats[MAX_PACKETS_PER_SECOND] )
	    {
		this.stats[MAX_PACKETS_PER_SECOND] = tmp;
	    }
	}
    }
    
    /**
     * Sampler is a helper class that is used to sample and recompute
     * certain statistics over a window of time.
     */
    private class Sampler
	implements Runnable
    {
	/**
	 * @param millis The number of milliseconds to sleep.
	 */
	public void sleep(long millis)
	{
	    try
	    {
		edu.utah.janos.nodeos.Thread.sleep(millis, 0);
	    }
	    catch(InterruptedException ie)
	    {
				/* Shouldn't ever happen... */
	    }
	}
	
	public void run()
	{
	    double oldStats[] = new double[STAT_MAX];
	    double newStats[] = new double[STAT_MAX];
	    long startTime, endTime, diffTime;
	    final int sampleTime = 333; // ms
	    double tmpStats[];
	    
	    System.arraycopy(NodeStat.this.stats,
			     0,
			     newStats,
			     0,
			     newStats.length);
	    while( true )
	    {
				/* Swap the old stats with the previous copy */
		tmpStats = oldStats;
		oldStats = newStats;
		newStats = tmpStats;
		startTime = NodeStat.this.time();
				/* Sleep for our time window */
		this.sleep(sampleTime);
		endTime = NodeStat.this.time();
		diffTime = endTime - startTime;
				/* Copy the current stat values */
		System.arraycopy(NodeStat.this.stats,
				 0,
				 newStats,
				 0,
				 newStats.length);
				/* Recompute any time dependent stats */
		NodeStat.this.recomputeStats(oldStats,
					     newStats,
					     endTime,
					     diffTime);
	    }
	}
    }
    
    NodeStat()
    {
	/* Record the startup time */
	this.stats[STARTUP_TIME] = (double)this.time();
	/* Start our sampler thread */
	Domain.currentDomain().threadStart(new Sampler(), "Stat sampler");
    }
    
    /**
     * @param index The statistic to retrieve.
     * @return The value of the specified statistic.
     */
    public double getStat(int index)
    {
	return this.stats[index];
    }
    
    /**
     * @param index The statistic to set.
     * @param value The value to give the statistic.
     */
    void setStat(int index, double value)
    {
	this.stats[index] = value;
    }
    
    /**
     * Increment the received packet count.
     */
    void receivedPacket()
    {
	this.stats[PACKETS_RECEIVED] += 1.0;
    }
    
    /**
     * Increment the sent packet count.
     */
    void sentPacket()
    {
	this.stats[PACKETS_SENT] += 1.0;
    }
}
