/*
 * PrimordialMessenger.java
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 */
package ants.core;

import ants.core.security.Principal;

import edu.utah.janosvm.kit.comm.commspace.CommQueuePuller;
import edu.utah.janosvm.kit.comm.commspace.CommQueuePusher;
import edu.utah.janosvm.kit.comm.commspace.CommSpaceElement;

/**
 * The PrimordialMessenger is used to send requests to the primordial domain.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
class PrimordialMessenger
    implements Runnable
{
    /**
     * The CommQueue the primordial domain will receive messages on.
     */
    private static final CommQueuePuller puller = new CommQueuePuller();
    
    /**
     * The CommQueuePusher used to send messages to the puller.
     */
    private static final CommQueuePusher pusher =
	new CommQueuePusher(puller);

    /**
     * The CommQueue the primordial domain will return messages on.
     */
    private static final CommQueuePuller retpuller = new CommQueuePuller();

    /**
     * The CommQuquePusher used to the send return messaegs on.
     */
    private static final CommQueuePusher retpusher =
	new CommQueuePusher(retpuller);
    
    /**
     * Add a MID to a running domain.
     */
    private final static int ADD_MID = 0;
    /**
     * Startup a protocol domain
     */
    private final static int INSTALL_PROTOCOL = 1;
    /**
     * Remove the application from the primordial domain tables.
     */
    private final static int DETACH_APPLICATION = 2;
    
    /**
     * Dispatch an AddMID message to the PrimordialNode
     */
    private void dispatchAddMID()
	throws InterruptedException
    {
	TypeID pid, mid, gid;
	UserID userInfo;
	int prevNode;
	
	pid = TypeID.wrap(puller.waitToTake().getByteArray());
	userInfo = new UserID(puller.waitToTake());
	mid = TypeID.wrap(puller.waitToTake().getByteArray());
	gid = TypeID.wrap(puller.waitToTake().getByteArray());
	prevNode = puller.waitToTake().getInt();
	PrimordialNode.addMID(pid, userInfo, mid, gid, prevNode);
    }
    
    /**
     * Dispatch an installProtocol message to the PrimordialNode
     */
    private void dispatchInstallProtocol()
	throws InterruptedException
    {
	TypeID pid;
	String name;
	Principal principal;
	
	pid = TypeID.wrap(puller.waitToTake().getByteArray());
	name = puller.waitToTake().getString();
	principal = new Principal(puller.waitToTake().getString());
	PrimordialNode.installProtocol(pid, name, principal);
	
	// Send return message
	CommSpaceElement retmsg[] = {
	    new CommSpaceElement(INSTALL_PROTOCOL)
	};
	retpusher.enqueue(retmsg);
    }
    
    /**
     * Dispatch an detachApplication message to the PrimordialNode
     */
    private void dispatchDetachApplication()
	throws InterruptedException
    {
	int port;
	
	port = puller.waitToTake().getInt();
	PrimordialNode.detachApplication(port);
    }
    
    /**
     * The message looper for the messenger.
     */
    public void run()
    {
	int kind;
	
	while( true )
	{
	    try
	    {
		/* Wait for the message kind */
		kind = puller.waitToTake().getInt();
		/* Dispatch the message */
		switch( kind )
		{
		case ADD_MID:
		    this.dispatchAddMID();
		    break;
		case INSTALL_PROTOCOL:
		    this.dispatchInstallProtocol();
		    break;
		case DETACH_APPLICATION:
		    this.dispatchDetachApplication();
		    break;
		}
	    }
	    catch(InterruptedException ie)
	    {
	    }
	}
    }
    
    /**
     * Send an addMID message to the PrimordialNode
     */
    public static void addMID(TypeID pid,
			      UserID userInfo,
			      TypeID mid,
			      TypeID gid,
			      int prevNode)
    {
	CommSpaceElement msg[] = {
	    new CommSpaceElement(ADD_MID),
	    new CommSpaceElement(pid.unwrap()),
	    userInfo.toCommSpaceElement(),
	    new CommSpaceElement(mid.unwrap()),
	    new CommSpaceElement(gid.unwrap()),
	    new CommSpaceElement(prevNode)
	};
	
	pusher.enqueue(msg);
    }
    
    /**
     * Send an installProtocol message to the PrimordialNode
     * <p>
     * This function blocks until the PrimordialNode has finished installing
     * this protocol
     */
    public static void installProtocol(TypeID pid, String name, Principal principal)
    {
	CommSpaceElement msg[] = {
	    new CommSpaceElement(INSTALL_PROTOCOL),
	    new CommSpaceElement(pid.unwrap()),
	    new CommSpaceElement(name),
	    new CommSpaceElement(principal.getName())
	};
	
	pusher.enqueue(msg);

	// Wait for the return message
	try
	{
	    retpuller.waitToTake();
	} catch (InterruptedException _) {
	    // This is probably for our own good
	}
    }
    
    /**
     * Send an detachApplication message to the PrimordialNode
     */
    public static void detachApplication(int port)
    {
	CommSpaceElement msg[] = {
	    new CommSpaceElement(DETACH_APPLICATION),
	    new CommSpaceElement(port)
	};
	
	pusher.enqueue(msg);
    }
}
