/* 
 * Protocol.java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package ants.core;

import edu.utah.janos.nodeos.Domain;

import ants.util.ByteCodeReader;
import ants.util.MD5;

import ants.core.security.Principal;

import java.util.Vector;

/**
 * Base class for all protocols. Intended to be subclassed for each
 * user-defined protocol for the purpose of collecting capsules into a
 * protocol and registering it at the local node.  A Protocol object
 * is a short-lived object used only for defining a new protocol.
 * After the protocol is defined and registered with the node it may
 * be thrown away.  This class could more effectively be called
 * "ProtocolDescription".
 * <p>
 * A protocol is organized into a collection of CodeGroups where each
 * CodeGroup contains one or more CapsuleClasses.  Unlike the Protocol
 * object, CodeGroup and CapsuleClass objects are long-lived and must
 * be kept in a cache on a node for the node to be able to process
 * the protocol.
 * <p>
 * A protocol is intially registered at an end node.  Classes are
 * loaded from the local file system to define the protocol.  When a
 * capsule is sent from a protocol, the receiving node (probably)
 * won't have the code to process it.  The receiving node will then
 * send a request for a specific CapsuleClass to the node which
 * sent the capsule.  The node will respond with the relevant
 * <emph>CodeGroup</emph> which contains the requested class.
 * Note that the receving node will never ask for a Protocol object
 * it just uses CodeGroups and CapsuleClasses.  There is no useful
 * state in a Protocol object.
 * <p>
 * This class should probably be abstract, but its used as a concrete
 * class for weak reasons in a number of test cases.
 * <p>
 * To define a protocol requires definining the relevant CodeGroups
 * and the CapsuleClasses contained in each CodeGroup.  For example:
 * <pre>
 * public class MulticastProtocol extends Protocol
 * {
 *   public MulticastProtocol()
 *   {
 *     startProtocolDefn();
 *  
 *     startGroupDefn();
 *     addCapsule("apps.MulticastCapsule");
 *     endGroupDefn();
 *  
 *     startGroupDefn();
 *     addCapsule("apps.MulticastSubscribeCapsule");
 *     endGroupDefn();
 *  
 *     endProtocolDefn();
 *   }
 * }
 * </pre>
 * XXX This class defines a new protocol with two code groups....
 *
 * @see CodeGroup
 * @see CapsuleClass
 * @author David Wetherall
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class Protocol
{
    /**
     * A list of CodeGroupData objects associated with this Protocol.
     * <p>
     * XXX Drop these vectors completely.  Just require a parameter
     * on StartProtocolDefn() that says how many groups will be defined.
     * (use arrays).
     */
    Vector groups = new Vector();
    
    /**
     * Temporary place to store capsule class data.  This is needed
     * to compute the methodID for a CapsuleClass.  CapsuleClassData
     * objects are discarded along with the Protocol object when
     * the protocol definition is complete.
     */
    static class CapsuleClassData
    {
	/**
	 * The name of the class this CapsuleClass represents.
	 */
	String name;
	
	/**
	 * The complete Class File this CapsuleClass represents.
	 */
	byte[] code;
	
	/**
	 * An MD5 hash of the Class File.
	 * <p>
	 * Except for builtin protocols, where this is
	 * a hash over the VERSION string in the
	 * Protocol.
	 *
	 * @see #code
	 */
	TypeID codeID;
	
	CapsuleClassData(String name, byte[] code, TypeID codeID)
	{
	    this.name = name;
	    this.code = code;
	    this.codeID = codeID;
	}
	
	/**
	 * Compute the hash code.
	 */
	CapsuleClassData(String name, byte[] code)
	{
	    this(name, code, TypeID.wrap(MD5.hash(code)));
	}
    }
    
    
    static class CodeGroupData
    {
	/**
	 * A list of CapsuleClassData objects
	 */
	final Vector classes = new Vector();
	
	TypeID groupID;
	
	/**
	 * Active classes have inchannels created for them.
	 * Non-active (inert) classes don't.  The only inert
	 * classes are in DLProtocol.
	 * <p>
	 * Technically, this should be a field in CapsuleClass,
	 * but that's a big waste of space, because there
	 * are only 3 inert capsules on a Node.
	 */
	boolean active;
	
	CodeGroupData(boolean active)
	{
	    this.active = active;
	}
    }
    
    /**
     * Start the definition of a new protocol.  Subclasses of Protocol
     * should invoke startProtocolDefn(), followed by
     * startGroupDefn()/endGroupDefn() pairs and finally
     * endProtocolDefn().  Calls to addCapsule() must be
     * inserted in the start/end group pairs.
     *
     * @see #endProtocolDefn()
     * @see #startGroupDefn()
     * @see #endGroupDefn()
     * @see #addCapsule(java.lang.String)
     */
    protected void startProtocolDefn()
    { 
	// Doesn't actually do anything at this point.
    }
    
    /**
     * End the protocol definition.  This ties all of the elements of
     * the protocol together.  Each CodeGroup is given a list
     * of all the other code Groups.
     *
     * @see #startProtocolDefn()
     * @see #startGroupDefn()
     * @see #endGroupDefn()
     * @see #addCapsule(java.lang.String)
     *
     * @exception Exception if the CodeGroup has too many chunks
     * (currently the limit is 32).
     */
    protected void endProtocolDefn() 
	throws Exception, ClassNotFoundException
    {
	if (false)
	    throw new Exception("This is stupid.");
	
    }
    
    /**
     * Activate this protocol on the given node.  Computes the
     * protocolID for this protocol, creates CodeGroup objects and
     * populates the CodeGroups with CapsuleClass objects.
     *
     * @param node the Node this Protocol is associated with.
     * @param register whether the protocol should be registered or fully
     * activated (adding InChannels and such).
     * @param principal the Principal the new protocol should start as.
     */
    void activate(Node node, boolean register, Principal principal)
	throws DuplicateCapsuleClassException,
	FullCodeGroupException,
	java.lang.ClassNotFoundException
    {
	// Compute the protocolID based on the groupIDs:
	TypeID[] groupIDs = new TypeID[groups.size()];
	for(int i = 0; i < groups.size(); i++)
	    groupIDs[i] = ((CodeGroupData)groups.elementAt(i)).groupID;
	
	TypeID protocolID = TypeID.computeFromIDs(groupIDs);
	
	/*
	 * If we're just registering the code we might have to start
	 * a new protocol domain
	 */
	if( register )
	{
	    PrimordialMessenger.installProtocol(protocolID,
						this.getClass().
						getName(),
						principal);
	}
	
	// At this point the protocol is completely defined.
	// Now we create and initialize the long-term data structures
	// that define the protocol.
	
	for(int i = 0; i < groups.size(); i++)
	{
	    CodeGroupData cgd = (CodeGroupData)groups.elementAt(i);
	    
	    // Skip CodeGroups that already exist in this Node.
	    // (this happens when apps start *after* protocols).
	    // XXX does it really?
	    if (node.lookupCodeGroup(cgd.groupID) != null)
		continue;
	    
	    // Create a group:
	    CodeGroup group = new CodeGroup(cgd.groupID,
					    groupIDs,
					    protocolID,
					    builtin());
	    
	    // Add capsules to the group:
	    for(int j = 0; j < cgd.classes.size(); j++)
	    {
		CapsuleClassData ccd = (CapsuleClassData)
		    cgd.classes.elementAt(j);
		TypeID methodID = CapsuleClass.computeMethodID(
		    ccd.codeID, cgd.groupID, protocolID);
		CapsuleClass cc =
		    new CapsuleClass(group,
				     methodID,
				     ccd.name,
				     ccd.code);
		group.add(cc);
	    }
	    
	    if( register )
		group.register(node);
	    else
		group.activate(node, cgd.active);
	}
	
	// Throw the list away, no one cares anymore.
	groups = null;
    }
    
    /**
     * Start a new CodeGroup definition.  Must be called in the
     * context of a ProtocolDefn().  Subsequent CapsuleClasses
     * will be added to this CodeGroup until it is ended.  Must
     * not be called after startGroupDefn() without an intervening
     * endGroupDefn().
     *
     * @see #startProtocolDefn()
     * @see #endProtocolDefn()
     * @see #endGroupDefn()
     * @see #addCapsule(java.lang.String)
     */
    protected void startGroupDefn() 
    { 
	startGroupDefn(true);
    }
    
    protected void startGroupDefn(boolean active)
    {
	CodeGroupData cgd = new CodeGroupData(active);
	groups.addElement(cgd);
    }
    
    /**
     * End a CodeGroup definition.  Must be called in
     * the context of a ProtocolDefn() and after an unmatched
     * startGroupDefn().  Ending a CodeGroup without having
     * added any capsules to it is a bad idea.
     *
     * @see #startProtocolDefn()
     * @see #endProtocolDefn()
     * @see #startGroupDefn()
     * @see #addCapsule(java.lang.String)
     */
    protected void endGroupDefn()
	throws Exception 
    {
	// XXX assert that some capsules were added to the code group.a
	
	// XXX assumes lastElement() is the lastElement added
	CodeGroupData g = (CodeGroupData)groups.lastElement();
	
	Vector classes = g.classes;
	TypeID[] codeIDs = new TypeID[classes.size()];
	for(int i = 0; i < classes.size(); i++)
	    codeIDs[i] = ((CapsuleClassData)classes.elementAt(i))
		.codeID;
	
	g.groupID = TypeID.computeFromIDs(codeIDs);
    }
    
    /**
     * Add a Capsule class to a CodeGroup.
     * Must be called in the context of a GroupDefn().
     * <p>
     * XXX rename to addCapsuleClass() or addCapsuleType()
     *
     * @param name Name of the class to include in the group.
     * @see #startProtocolDefn()
     * @see #endProtocolDefn()
     * @see #startGroupDefn()
     * @see #endGroupDefn()
     */
    protected void addCapsule(String name)
	throws Exception // XXX explicitly list exceptions
    {
	byte[] code = Domain.readByteCode(name);
	
	Assert.assert(code != null);

	CodeGroupData g = (CodeGroupData)groups.lastElement();
	g.classes.addElement(new CapsuleClassData(name, code));
    }
    
    /**
     * Returns non-zero if this defines a builtin protocol.
     * <p>
     * Overridden in BuiltinProtocol.
     *
     * @see BuiltinProtocol
     */
    boolean builtin()
    {
	return false;
    }
    
    /**
     * Return a string representation of this
     * Protocol.
     *
     * @return A string.
     */
    public String toString() 
    {
	return this.getClass().getName() + "["
	    + (builtin() ? "(builtin)" : "(dl'd)")
	    + "]";
    }
}
