/*
 * ProtocolBooster.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core;

import edu.utah.janos.nodeos.Address;
import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.InChannel;
import edu.utah.janos.io.Log;

import edu.utah.janosvm.kit.comm.commspace.CommQueuePusher;
import edu.utah.janosvm.kit.comm.commspace.CommQueuePuller;
import edu.utah.janosvm.kit.comm.commspace.CommSpaceHandle;
import edu.utah.janosvm.kit.comm.commspace.CommSpaceElement;

import java.io.IOException;

import ants.core.security.SecurityDefaults;
import ants.core.security.Principal;

/**
 * Must be public as its the entrypoint for a new domain from the
 * Java NodeOS.
 * <p>
 * This is invoked by PrimordialNode as the intial class in a new
 * domain.  This new domain is charged with taking the new (unknown
 * protocol) PID, requesting and loading the code associated with that
 * PID and then processing any enqueued packets.
 * <p>
 * Protocol's have a restricted namespace.  See the exportedAntsClasses
 * list in PrimordialNode.
 *
 * @see AppBooster
 * @see PrimordialNode#fireUpNewProtocolDomain(TypeID,CommQueuePusher)
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class ProtocolBooster
    extends Booster
{
    /**
     * The booster communicates with the Primordial Node through a
     * CommQueue so that it can figure out what protocols it is supposed
     * to be handling.  The booster can either install a new mid through
     * dynamic loading, or load up a locally defined protocol.  To
     * distinguish between them the first element pushed onto the queue
     * is an integer indicating what it is supposed to do, followed by
     * any needed data.
     */
    
    /**
     * Indicates a new MID, arguments are the MID bytes and the previous
     * node that the packet was at.
     */
    public final static int NEW_MID = 0;
    
    /**
     * Indicates a local protocol should be instantiated, the only argument
     * is the name of the protocol class.  Once this executes the protocol
     * message handler thread is basically useless since the whole
     * protocol is active.
     */
    public final static int LOCAL_PROTOCOL = 1;
    
    /**
     * Indicates an application capsule is queued up.  Application capsules
     * are queued here if they arrive before a new protocol is completely
     * set up.
     */
    public final static int APP_CAPSULE = 2;
    
    private boolean local = false;
    
    public ProtocolBooster(edu.utah.janos.nodeos.ClassInstantiationKey key)
	throws SecurityException
    {
	key.checkEquals(ProtocolBooster.class);
    }
    
    /**
     * The entrypoint invoked in the context of the new domain.
     */
    public void run(final Domain domain,
		    String[] args,
		    CommSpaceHandle commSpaceHandle)
    {
	final CommQueuePuller queue = new CommQueuePuller();
	BoosterArgs ba = new BoosterArgs(commSpaceHandle);
	final Node node;
	int lpc;
	
	Assert.require(domain);
	Assert.require(args != null);
	Assert.require(args.length == 0); // XXX not expecting any
	Assert.require(commSpaceHandle);
	
	domain.setName("Protocol <unknown>");
	Trace.trace(Trace.ANTS_DYNLOAD, "Starting unknown protocol. domain=" +domain);
	
	// Turn on the package-level access controls for this domain.
	domain.enableAccessControls(true);
	// Add any system packages and seal them up.
	for( lpc = 0; lpc < PrimordialNode.antsPackages.length; lpc++ )
	{
	    Domain.currentDomain().addSystemPackage(PrimordialNode.antsPackages[lpc]);
	    Domain.currentDomain().sealPackage(PrimordialNode.antsPackages[lpc]);
	}
	// Export the ANTS classes for the protocols to use.
	for( lpc = 0; lpc < PrimordialNode.exportedAntsClasses.length; lpc++ )
	{
	    Domain.currentDomain().exportClass(PrimordialNode.exportedAntsClasses[lpc]);
	}
	
	// setupLocalNode is in Booster superclass.
	node = setupLocalNode(domain, ba);

	/* Pass our message queue to the primordial domain */
	ba.setDomainMsgQueue(queue);
	// Dedicate a thread to handling new MIDs:
	domain.threadStart(new Runnable() {
	    public void run() {
		msgLoop(domain, queue, node);
	    }
	},
	    "MID/Protocol Handler for " + ba.getPID());
	
	Trace.trace(Trace.ANTS_DYNLOAD, "Unknown protocol on its way! domain=" +domain);
    }
    
    /**
     * This will wait for and process any messages coming in on the
     * CommQueue
     */
    private void msgLoop(Domain domain, CommQueuePuller queue, Node node)
    {
	while(!domain.destroyed())
	{
	    try
	    {
				/* Dispatch any messages we get */
		int message =
		    queue.waitToTake().getInt();
		this.dispatch(domain, message, queue, node);
	    }
	    catch(Exception e) // dispatch() throws 6 kinds
	    {
		node.log(Log.ERR,
			 "Exception starting new mid in "
			 + Domain.currentDomain()
			 + ": "
			 + e.getMessage());
		node.logStackTrace(Log.ERR, e);
	    }
	    catch(Throwable e)
	    {
		node.log(Log.ERR,
			 "Error starting new mid in "
			 + Domain.currentDomain()
			 + ": "
			 + e.getMessage());
		node.logStackTrace(Log.ERR, e);
	    }
	}
	Trace.trace(Trace.ANTS_DYNLOAD, "Protocol looper quitting");
    }
    
    /**
     * This will actually dispatch a message in the queue
     *
     * @param domain The current domain
     * @param message The integer identifier for the message
     * @param queue The queue the message was sent on
     * @param node The local node object
     */
    private void dispatch(Domain domain,
			  int message,
			  CommQueuePuller queue,
			  Node node)
	throws java.lang.InterruptedException,
	java.lang.ClassNotFoundException,
	java.lang.InstantiationException,
	java.lang.IllegalAccessException,
	FullCodeGroupException,
	DuplicateCapsuleClassException
    {
	switch(message)
	{
	case NEW_MID:
	    Trace.trace(Trace.ANTS_DYNLOAD, "ProtocolBooster.java: NEW_MID ");
	    byte[] midBytes =
		queue.waitToTake().getByteArray();
	    byte[] gidBytes =
		queue.waitToTake().getByteArray();
	    int prev = queue.waitToTake().getInt();
	    
	    if( local )
		break;
	    
	    // XXX this is a bit hokey.
	    if (node.getPrincipal() == SecurityDefaults.BOOTSTRAP_USER)
		node.setPrincipal(SecurityDefaults.REMOTE_USERS);

	    TypeID mid = TypeID.wrap(midBytes);
	    TypeID gid = TypeID.wrap(gidBytes);
	    
	    Trace.trace(Trace.ANTS_DYNLOAD, Domain.currentDomain()+ " got MID " +mid+ " through CommQueue.");
	    startNewGid(node, gid, mid, prev);
	    break;
	case LOCAL_PROTOCOL:
	    String protocolName = queue.waitToTake().getString();
	    Principal principal = new Principal(queue.waitToTake().getString());
	    Protocol pr;
	    Class cl;
	    String pkgName;
	    
	    Trace.trace(Trace.ANTS_DYNLOAD, "ProtocolBooster.java: LOCAL_PROTOCOL " +protocolName+ " in " +domain) ;

	    if( local )
		break;
	    
	    domain.setName("Protocol " +protocolName);
	    
	    // Allow the protocol to be loaded off the disk.
	    domain.exportClass(protocolName);
	    
	    Trace.trace(Trace.ANTS_DYNLOAD, "ProtocolBooster.java: LOCAL_PROTOCOL "
		+protocolName+ " principal is " +principal);

	    node.setPrincipal(principal);

	    /* Load and start it up */
	    Trace.trace(Trace.ANTS_DYNLOAD, Domain.currentDomain()+ " loading LOCAL_PROTOCOL " +protocolName);
	    cl = domain.loadClass(protocolName);
	    pr = (Protocol)cl.newInstance();
	    pr.activate(node, false, principal);
	    this.local = true;
	    Trace.trace(Trace.ANTS_DYNLOAD, Domain.currentDomain()+ " loading LOCAL_PROTOCOL " +protocolName);
	    break;
	case APP_CAPSULE:
	    Capsule cap = Capsule.dequeueCapsule(node, queue);
	    CommQueuePusher replyQueue = queue.waitToTake().
		getCommQueue();
	    boolean sent;

	    Trace.trace(Trace.ANTS_DYNLOAD, "ProtocolBooster.java: APP_CAPSULE in " +domain);
	    
	    // XXX should dequeueCapsule really be returning null?
	    if (cap == null)
		sent = false;
	    else 
		sent = cap.evaluate(node);

	    // XXX reply is not sent if there is an exception above
	    replyQueue.enqueue(new CommSpaceElement(sent ? 1 : 0));
	    break;
	default:
	    node.log(Log.ERR,
		     "Invalid message sent to protocol "
		     + "message handler");
	    break;
	}
    }
    
    /**
     * Start the process of installing the new protocol's code.
     * This installs a channel for matching DLResponseCapsules
     * that contain code for this protocol, then it sends off
     * a DLRequestCapsule to get the code.
     * <p>
     * It's possible for a MID to get started twice.  But this
     * is probably ok, and maybe even necessary: if the first
     * installation fails halfway through, a second installation
     * has to complete it.
     */
    private void startNewGid(Node node, TypeID gid,
			     TypeID mid, int prev)
    {
	// Setup a Key for receving DLReply packets...
	new CapsuleChannel(DLResponseCapsule.class,
			   new CodeRequestDemultiplexKey(
							 Domain.currentDomain(),
							 DLResponseCapsule.MID,
							 gid.unwrap()));
	
	// Construct a DLRequest packet and send it to prevNode.
	DLRequestCapsule req = new DLRequestCapsule(prev, mid);
	req.prime(node.getAddress(), 1);
	node.routeForNode(req, prev);
	/* Add a pending request for this DL request */
	node.getCache().put(gid,
			    new PendingRequest(req),
			    2);
    }
}
