/*
 * UnknownPacketHandler.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core;

import edu.utah.janos.nodeos.BufferHandle;
import edu.utah.janos.nodeos.DeliverFunc;
import edu.utah.janos.nodeos.DemultiplexKey;
import edu.utah.janos.nodeos.Domain;
import edu.utah.janos.nodeos.InChannel;
import edu.utah.janos.nodeos.ProtocolSpecFormatException;
import edu.utah.janos.nodeos.PacketAddressFormatException;

import java.io.IOException;

/**
 * Handle an ANTS packet of unknown MID
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
final class UnknownPacketHandler
    implements DeliverFunc
{
    private final Node node;
    private final DemultiplexKey classifier;
    private final java.util.Random random = new java.util.Random(); // XXX
    private final InChannel inChan;
    
    private static final float RANDOM_DUPMIDREQ_FRACTION = 0.05F;
    
    /**
     * Create the UnknownPacketHandler.  Only need one of these
     * per node...
     */
    /*package*/ UnknownPacketHandler(Node node)
    {
	this(node, new ANEPDemultiplexKey(Domain.currentDomain(),
	    				  Capsule.getAnepTypeID()));
    }
    
    /*package*/ UnknownPacketHandler(Node node, DemultiplexKey classifier)
    {
	this.node = node;
	this.classifier = classifier;
	try
	{
	    this.inChan = InChannel.create(Domain.currentDomain(),
					   this.classifier,
					   PrimordialNode.getReceiveProtocolSpec(),
					   PrimordialNode.getReceiveAddressSpec(),
					   this);
	    
	    Trace.trace(Trace.ANTS_DYNLOAD,
			"Created InChan for unknown packets: "
			+ this.inChan);
	    inChan.addBuffers(30);
	}
	catch(PacketAddressFormatException e)
	{
	    Node.abort(e, "Bad physical address spec");
	    throw new InternalError("this won't ever happen");
	}
	catch(ProtocolSpecFormatException e)
	{
	    Node.abort(e, "Bad physical protocol spec");
	    throw new InternalError("this won't ever happen");
	}
	catch(IOException e)
	{
	    Node.abort(e, "IOException when creating in channel " + e);
	    throw new InternalError("this won't ever happen");
	}
    }
    
    /**
     * Unknown packet handler.  For ANTS packets that don't match
     * an installed MID.
     * <p>
     * This is the only method defined by the DeliverFunc interface.
     *
     * @see DeliverFunc
     */
    public final void receive(BufferHandle bufh, int payloadStart)
    {
	Trace.trace(Trace.ANTS_DYNLOAD, "Received unknown ANTS packet " +bufh.toString()
		    + " w/payloadStart=" + payloadStart);
	
	if( bufh.getLength() < (payloadStart + Xdr.INT + Xdr.TYPEID) )
	{
	    // XXX log this event.
	    Trace.trace(Trace.ANTS_DYNLOAD, "Unknown packet is bad: length < " +(payloadStart + Xdr.INT + Xdr.TYPEID)+ " bytes");
	    
	    // Recycle the bufh
	    inChan.addBuffer(bufh);
	    
	    return;
	}
	
	/*
	 * By definition this receive() function only
	 * gets packets for an unknown ANTS protocol.
	 */
	
	byte[] midbuf = new byte[Xdr.TYPEID];
	bufh.getBytes(payloadStart + Xdr.INT, midbuf, 0, Xdr.TYPEID);
	TypeID mid = TypeID.wrap(midbuf).intern();
	Assert.assert(mid.intern() == mid);
	
	// XXX don't know what will happen with bufh (most
	// likely will get queued up in th unknownMidPackets map)
	// So add a replacement buffer to the inChan
	inChan.addBuffer(BufferHandle.create());
	
	// XXX create a new BufferHandle that subarrays from payloadStart on?
	payloadStart += Xdr.INT + Xdr.TYPEID;
	
	Trace.trace(Trace.ANTS_DYNLOAD, "handling unknown mid (" +mid+ ")");
	
	// XXX loses payloadstart...
	boolean midAlreadySeen = PrimordialNode.unknownMidPackets
	    .addPacket(mid.unwrap(), bufh);
	
	if(midAlreadySeen)
	{
	    /*
	     * This mid request has already been sent so we don't
	     * have to worry about it since the pending request
	     * mechanism "should" take care of it.
	     */
	    Trace.trace(Trace.ANTS_DYNLOAD, " mid (" +mid+ ") is already known.");
	    return;
	}
	
	
	/*
	 * XXX this is hackish: For convenience, pull out the
	 * address of the node that this packet came from.
	 * The new domain will need the info for firing off a
	 * DLRequest capsule...
	 *
	 * Format after mid is:
	 * src (int), dst (int), and previous (int)
	 */
	int prev = bufh.getInt(payloadStart + 4 + 4);
	
	int myAddr = node.getAddress();
	if (myAddr != 0)
	{
	    MIDIDRequestCapsule reqCap;

	    Trace.trace(Trace.ANTS_DYNLOAD,
		"Sending MIDIDRequest for mid (" + mid + ") to "
		+NodeAddress.toString(prev));

	    reqCap = new MIDIDRequestCapsule(prev, mid);
	
	    /*
	     * Setup the req packet to go back to whence it
	     * came.
	     */
	    reqCap.prime(myAddr, 1);
	    
	    node.routeForNode(reqCap, prev);
	    /* Add a pending request for this mid request */
	    node.getCache().put(mid,
		new PendingRequest(reqCap),
		2);
	}
	else
	{
	    /*
	     * XXX what do we do in this case?  Queue the packet up to deal with
	     * it once we get an address?
	     */
	    Trace.trace(Trace.ANTS_DYNLOAD,
		"Cannot reply to packet, as I have no address!");
	}
    }
    
    public String toString()
    {
	return "UnknownPacketHandler for " +node;
    }
}
