/* 
 * DanteServer.java
 *
 * Copyright (c) 2000, 2001 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.dante;

import ants.core.*;
import ants.util.*;
import edu.utah.janos.nodeos.Address;
import edu.utah.janos.nodeos.AddressFactory;
import edu.utah.janos.nodeos.NoSuchNodeAddress;
import edu.utah.janos.nodeos.AddressFormatException;

import java.util.*;

/** 
 * This class dynamically distributes addresses for a core server.
 * <p>
 * DanteServer accepts the following arguments:<br>
 * <menu>
 * <li> <b>-port  <var>int</var></b>:     override the default port value (14)</li>
 * <li> <b>-start <var>X.Y.Z.W</var></b>: the first address available for distribution </li>
 * <li> <b>-end   <var>A.B.C.D</var></b>: the last address available for distribution</li>
 * </menu>
 *
 * @author Andrew Whitaker
 */
public final class DanteServer extends Application 
    implements Runnable, DanteConstants
{
    private static final AddressFactory af = new AddressFactory();
    
    private static final String[] defaults = 
    {"-port", Integer.toString(DANTE_SERVER_PORT)};
    
    /** The first address managed by the server */
    private int firstAddress;
    
    /** The last address managed by the server */
    
    private int lastAddress;
    
    /** The number of addresses managed by the server */
    private int n;
    
    /** The number of allocated addresses */
    private int allocated;
    
    /** A vector of addresses currently in use.  The hash values 
	correspond to boolean use bits */
    private Hashtable addresses;
    
    /** The local node object */
    private Node node;
    
    /** A table of neighboring nodes */
    private NeighborTable neighbors;
    
    /** The first available address (probably) */
    private int index;
    
    public DanteServer () {
	addresses = new Hashtable();
	index = 0;
	allocated = 0;
	firstAddress = lastAddress = 0;
    }
    
    public void start () throws Exception {
	node = this.getNode();
	neighbors = node.getNeighborTable();
	
	node.register(new DanteProtocol(), getPrincipal());
	node.threadStart(this);
    }
    
    public void setArgs(KeyArgs k) 
	throws InvalidKeyArgsException
    {
	k.merge(defaults);
	
	for (int i = 0; i < k.length(); i++) {
	    if (k.key(i).equals("-start")) {
		try {
		    firstAddress = Address.fromString(k.arg(i));
		    k.strike(i);			
		}
		catch (AddressFormatException afe) {
		    throw new InvalidKeyArgsException 
			("Address format exception: " + k.arg(i));
		}
	    }
	    
	    else if (k.key(i).equals("-end")) {
		try {
		    lastAddress = Address.fromString(k.arg(i));
		    k.strike(i);			
		}
		catch (AddressFormatException afe) {
		    throw new InvalidKeyArgsException 
			("Address format exception: " + k.arg(i));
		}
	    }		
	}
	
	if (lastAddress < firstAddress)
	    throw new InvalidKeyArgsException 
		("The ending address must be larger than the start!");
	
	n = (lastAddress - firstAddress) + 1;
	super.setArgs(k);
    }
    
    public void receive (Capsule c) {
	if (! (c instanceof DanteCapsule))
	    return;
	DanteCapsule cap = (DanteCapsule) c;
	
	int type = cap.getType();
	Address link = null;
	
	try {
	    link = af.newAddress(cap.getEdgeVNA(),
				 cap.getEdgeProtocol(),
				 cap.getEdgePhysAddress());
	}
	catch (NoSuchNodeAddress nsna) {
	    Trace.trace(Trace.ANTS_DANTE, "DanteServer: No such node address" +
			cap.getEdgeProtocol() + "  " + cap.getEdgePhysAddress());
	    return;
	}
	
	switch(type) {
	case JOIN: {
	    try {
		int address = allocateAddress(link);
		Trace.trace(Trace.ANTS_DANTE, "DanteServer: Allocated address: "
			    +Address.toString(address));
		
		cap.setEdgeVNA(address);
		cap.setCoreVNA(node.getAddress());
		cap.setType(DanteCapsule.RESPONSE);
		node.sendToNeighbor(cap,address);
	    }
	    catch (AddressExhaustionException ahe) {
		Trace.trace(Trace.ANTS_DANTE, "DanteServer: Addresses are exhausted!");
		// drop the capsule
	    }
	    break;
	}
	
	case REFRESH: {
	    Trace.trace(Trace.ANTS_DANTE, "DanteServer: Keep alive received for: "
			+Address.toString(cap.getEdgeVNA()));
	    markAddress (cap.getEdgeVNA(),link);
	    break;
	}
	case LEAVE: {
	    Trace.trace(Trace.ANTS_DANTE, "DanteServer: Edge node leaving: "
			+Address.toString(cap.getEdgeVNA()));	 
	    returnAddress(cap.getEdgeVNA());
	    break;
	}	
	default: {
	    Trace.trace(Trace.ANTS_DANTE, "DanteServer: Unknown capsule type.");	
	}
	}
    }
    
    /** 
     * Allocate a new (unused) address, and a build a provisional gateway  
     */
    public synchronized int allocateAddress (Address link) 
	throws AddressExhaustionException 
    {
	
	if (allocated == n)
	    throw new AddressExhaustionException();
	
	while (true) {
	    Integer key = new Integer (index);
	    Boolean b = (Boolean)addresses.get(key);
	    if (b == null) {
		allocated++;
		addresses.put(key,Boolean.FALSE);
		
		int address = firstAddress + index;
		neighbors.addNeighbor(address,link);
		
		index = (index + 1) % n;
		return address;
	    }
	    else index = (index + 1) % n;
	}
    }
    
    /** 
     * Called on receiving a keep-alive update to reset the timeout field.
     * This prevents the address from being garbage collected.  Also, add
     * the address to the list of gateways if it isn't already there.  
     */      
    public synchronized void markAddress (int address,Address link) {
	
	int index = address - firstAddress;
	Integer key = new Integer (index);
	Boolean b = (Boolean) addresses.get(key);
	if (b == Boolean.FALSE) {
	    addresses.put(key,Boolean.TRUE);
	    
	    // Make sure this gateway is in the table 
	    if (neighbors.getNeighbor(address) == null)
		neighbors.addNeighbor (address,link);
	}
    }
    
    /** 
     * Return a timed-out address to the pool. 
     */
    public synchronized void returnAddress(int address) {
	Trace.trace(Trace.ANTS_DANTE, "DanteServer: Edge node returning: "
		    +Address.toString(address));
	
	int index = address - firstAddress;
	Integer key = new Integer (index);
	
	Object o = addresses.get(key);
	if (o != null) {
	    allocated--;
	    addresses.remove(key);
	    
	    // Remove this gateway
	    neighbors.removeNeighbor(address);
	}
    }
    
    /** 
     * Clear the use bits after each timeout period.  Discard old
     * address mappings.
     */
    
    public void run () {
	while (true) { 
	    try { 
		Node n = Node.currentNode();
		n.sleep(TIMEOUT); 
	    } 
	    catch (InterruptedException _) {}
	    
	    synchronized(this) {
		
		Enumeration e = addresses.keys();
		while (e.hasMoreElements()) {
		    Integer key = (Integer) e.nextElement();
		    Boolean b = (Boolean)addresses.get(key);
		    
		    if (b == Boolean.TRUE) {
			addresses.put(key,Boolean.FALSE);
		    }
		    else {
			returnAddress (key.intValue() + firstAddress);
		    }	  
		}
	    }
	}
    }
    
}
