/*
 * DynamicRouteEntry.Java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * Copyright (c) 2000 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.route;

import edu.utah.janos.nodeos.NoSuchNodeAddress;
import edu.utah.janos.nodeos.AddressFormatException;
import edu.utah.janos.nodeos.Address;

import edu.utah.janos.route.RouteEntry;

import ants.core.Assert;
import ants.core.Trace;
import ants.core.RouteTable;

/**
 * A DynamicRouteEntry is a RouteEntry with weight.
 * <p>
 * Unlike a base RouteEntry, which is immutable, a 
 * DynamicRouteEntry's cost, expires and change fields
 * are mutable.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 * @author <a href="mailto:djw@cs.washington.edu">David Wetherall</a>
 */
class DynamicRouteEntry
    extends RouteEntry
{
    // Cost stuff used by the RouteTable update code.
    private Address addr;
    private int cost;
    private long expires;
    boolean change; // XXX make private, force access through methods
    
    public DynamicRouteEntry(int dest, int nxt, String physAddr)
	throws NumberFormatException, AddressFormatException, NoSuchNodeAddress
    {
	this(dest,RouteTable.parseAddress(nxt, physAddr));
    }
    
    public DynamicRouteEntry(int dest, Address physAddr)
    {
	super(dest, physAddr);
	this.addr = physAddr;
	change = false;
	cost = -1; // XXX
	expires = -1; // XXX
    }
    
    public DynamicRouteEntry(DynamicRouteEntry other)
    {
	super(other);
	this.addr = other.addr;
	change = other.change;
	cost = other.cost;
	expires = other.expires;
    }

    public DynamicRouteEntry(RouteEntry re)
    {
	super(re);
	this.addr = (Address)re.getNextHop();
	change = true;
	cost = -1;
	expires = -1;
    }
    
    public int getCost()
    {
	Assert.assert(cost > 0);
	return cost;
    }
    
    public void setCost(int cost)
    {
	this.cost = cost;
    }
    
    public long getExpiration()
    {
	// Assert.assert(expires > 0);
	return expires;
    }
    
    public void setExpiration(long expires)
    {
	Assert.require(expires > 0);
	this.expires = expires;
    }
    
    public boolean changed()
    {
	return change;
    }
    
    public void markChanged()
    {
    }
    
    public void clearChanged()
    {
    }

    public int getLogicalDest()
    {
	return super.getAddress();
    }

    public int getLogicalNext()
    {
	return this.addr.getANAddress();
    }

    public Address getPhysicalNext()
    {
	return this.addr;
    }
    
    public String toString() {
	return "DynamicRouteEntry["
	    + "d=" +Address.toString(getLogicalDest()) + ";"
	    + "n=" +Address.toString(getLogicalNext()) + ";"
	    + "a=" +getPhysicalNext() + ";"
	    + "x=" +getExpiration() + ";"
	    + "$=" +getCost()
	    + "]";
    }
}
