/*
 * DynamicRouteEvents.Java
 * 
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * Copyright (c) 2000 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package ants.route;

import java.util.Vector;
import java.util.Enumeration;

import ants.core.Trace;
import ants.core.Assert;

/**
 * Event queue for the dynamic route table.  Also contains
 * all of the timeout-related constants.
 */
class DynamicRouteEvents
{
    /**
     * eventTime() rounds all timers to this granularity
     * (in milliseconds).
     *
     * @see #eventTime(long,long,long)
     */
    private final static long RCRESOLUTION = 1000;
    
    // some RIP-like constants
    private final static int SECONDS = 1000;
    private final static int EXPIREINTERVAL = 60 * SECONDS;
    private final static int GCINTERVAL = 30 * SECONDS;
    private final static int UPDATEINTERVAL = 7 * SECONDS;
    private final static int UPDATESPREAD = 2 * SECONDS;
    private final static int TRIGGERINTERVAL = 2 * SECONDS;
    private final static int TRIGGERSPREAD = 1 * SECONDS;
    
    /**
     * Store of timer events.  Each event is a DynamicRouteEvent.
     * Not sorted or in any particular order (probably never
     * long enough to justify).  There is exactly one
     * event set to expire for any given time.
     * <p>
     * Used for expire, GC and periodic timers.
     *
     * @see #findTimer(long)
     * @see #findFirstTimer()
     * @see #setTimer(long,int)
     * @see #clearTimer(DynamicRouteEvent)
     */
    private final Vector events;
    
    DynamicRouteEvents()
    {
	this.events = new Vector(16);
    }
    
    boolean isEmpty()
    {
	return events.isEmpty();
    }
    
    /**
     * Compute a route event clock time.  All returned times
     * are snapped to granularity of RCRESOLUTION.
     *
     * @param base Generally 'now'.
     * @param add offset to add to base
     * @param spread for randomizing the time a bit.
     */
    static long eventTime(long base, long add, long spread) 
    {
	long time = base + add + (long)((double)spread*(Math.random()-0.5));
	Assert.assert(time > base);
	time += RCRESOLUTION - 1 ;
	time -= (time % RCRESOLUTION);
	return time;
    }
    
    /**
     * Find a DynamicRouteEvent set to go off at "when".
     * <p>
     * Finds the first such event.
     * <p>
     * There is only ever be one (or zero) DynamicRouteEvent in the queue
     * with a timer that matches when.
     */
    DynamicRouteEvent findTimer(long when)
    {
	Assert.assert(when > 0, "when must be greater than zero (when="+when+")");
	
	for (Enumeration e = events.elements(); e.hasMoreElements();)
	{
	    DynamicRouteEvent r = (DynamicRouteEvent)e.nextElement();
	    if (r.when == when)
	    {
		return r;
	    }
	}
	return null;
    }
    
    /**
     * Find the event in the timer queue with the shortest
     * timer.
     */
    public DynamicRouteEvent findFirstTimer()
    {
	if (events.isEmpty())
	    return null;
	
	DynamicRouteEvent low = (DynamicRouteEvent)events.firstElement();
	for (Enumeration e = events.elements(); e.hasMoreElements();)
	{
	    DynamicRouteEvent r = (DynamicRouteEvent)e.nextElement();
	    if (r.when < low.when) low = r;
	}
	
	return low;
    }
    
    /**
     * Add a new timer to the event queue.
     * <p>
     * Reuses existing timers if one is set to go off at the same time.
     */
    public DynamicRouteEvent setTimer(long when, int type)
    {
	DynamicRouteEvent re = findTimer(when);
	
	if (re == null)
	{
	    re = new DynamicRouteEvent(when, type);
	    events.addElement(re);
	}
	else
	{
	    re.what |= type;
	}
	
	Trace.trace(Trace.ANTS_ROUTE_TIMER, "Added timer, now " + events.size() + " events pending");
	
	return re;
    }
    
    /**
     * Remove the given DynamicRouteEvent from the list.
     *
     * @param r the event to remove
     */
    public void clearTimer(DynamicRouteEvent r)
    {
	events.removeElement(r);
	
	Trace.trace(Trace.ANTS_ROUTE_TIMER, "Removed timer, now " + events.size() + " events remain");
    }
    
    /**
     * Add a expire timer to go off after a constant EXPIREINTERVAL.
     * XXX can see RouteEntry.
     *
     * @see #EXPIREINTERVAL
     */
    public void setExpireTimer(DynamicRouteEntry r, long now)
    {
	long when = eventTime(now, EXPIREINTERVAL, 0);
	r.setExpiration(when);
	DynamicRouteEvent re = setTimer(when, DynamicRouteEvent.EXPIRE);
	
	// Track the number of routes that will be expired at this time
	re.expireCount++;
    }
    
    /**
     * Clear a expire timer from the event queue.  The event
     * found at 'when' had best be a expire event with a non-zero
     * expire count, thank you very much.
     *
     * @param when when the timer was scheduled to go off.
     */
    public void clearExpireTimer(long when)
    {
	DynamicRouteEvent re;

	Assert.require(when > 0);

	re = findTimer(when);
	
	// re must be a EXPIRE event
	if ((re==null) || ((re.what&DynamicRouteEvent.EXPIRE)==0) || (re.expireCount==0))
	    throw new InternalError("Bad reference count implementation for EXPIRE. re==" +re+ " @ " +when);
	
	// If the ref count goes to zero, then remove the event entirely.
	if (--re.expireCount == 0)
	    re.what = re.what & (~DynamicRouteEvent.EXPIRE);
	
	// If the event isn't for anything anymore, expire the whole thing.
	if (re.what == DynamicRouteEvent.EMPTY)
	    clearTimer(re);
    }
    
    /**
     * Add a GC timer to go off a constant GCINTERVAL from now.
     * Update the given route to 'expire' at that time.
     * 
     * XXX can see RouteEntry.
     */
    public void setGCTimer(DynamicRouteEntry route, long now)
    {
	long when = eventTime(now, GCINTERVAL, 0);
	route.setExpiration(when);
	DynamicRouteEvent re = setTimer(when, DynamicRouteEvent.GC);
	re.gcCount++;
    }
    
    /**
     * Clear a GC timer from the event queue.  There had
     * best be a GC timer at moment when.
     */
    public void clearGCTimer(long when)
    {
	DynamicRouteEvent re = findTimer(when);
	
	// GC timer had best exist
	if ((re == null) || ((re.what & DynamicRouteEvent.GC)==0) || (re.gcCount == 0))
	    throw new InternalError("Bad reference count implementation for DynamicRouteEvents.GC");
	
	// If gcCount goes to zero, then drop it has a GC event
	if (--re.gcCount == 0)
	    re.what = re.what & (~DynamicRouteEvent.GC);
	
	// If event has no purpose, remove it from queue.
	if (re.what == DynamicRouteEvent.EMPTY)
	    clearTimer(re);
    }
    
    /**
     * Add a periodic timer to go off at UPDATEINTERVAL +/- UPDATESPREAD
     * millis after 'now'.
     *
     * @see #UPDATESPREAD
     * @see #UPDATEINTERVAL
     */
    public void setPeriodicTimer(long now)
    {
	long when = eventTime(now, UPDATEINTERVAL, UPDATESPREAD);
	setTimer(when, DynamicRouteEvent.PERIODIC);
    }
    
    /**
     * Add a trigger timer to go off at TRIGGERINTERVAL +/- TRIGGERSPREAD
     * millis after 'base'.
     * <p>
     * Sets the 'trigger' member to true.
     *
     * @see #TRIGGERINTERVAL
     * @see #TRIGGERSPREAD
     */
    public void setTriggerTimer(long base)
    {
	long when = eventTime(base, TRIGGERINTERVAL, TRIGGERSPREAD);
	setTimer(when, DynamicRouteEvent.TRIGGER);
    }
}
