/*
 * DynamicRouteTable.Java
 * 
 * Copyright (c) 1999, 2001, 2002 The University of Utah and the Flux Group.
 * Copyright (c) 2000 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package ants.route;

import ants.core.Node;

import edu.utah.janos.io.Log;

import edu.utah.janos.route.RouteEntry;

import java.io.IOException; // XXX

import java.util.Hashtable; // XXX
import java.util.Enumeration; // XXX

import ants.core.Trace;
import ants.core.Assert;
import ants.core.NeighborTable;
import ants.core.RouteTable;

/**
 * The ANTS DynamicRouteTable contains all of the state and code
 * for maintaining a dynamic routing table.  This includes
 * sending out periodic updates, sending out immediate updates,
 * (and receiving both).  GC'ing stale routes, etc.
 * <p>
 * I think there are some problems in here for a node that has
 * more than one physical address.  It will probably send the
 * wrong info out to the other nodes...
 * <p>
 * Based very strongly on the dynamic route table from the original
 * MIT ANTS.  I renamed and moved a lot of things around, but
 * I (hopefully) didn't change the update algorithm.
 * <p>
 * XXX odd interaction with the base routetable
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 * @author David Wetherall
 */
public class DynamicRouteTable
{
    /**
     * The base route table in the ANTS runtime that
     * this application will add to.
     */
    private final RouteTable baseRouteTable;

    /**
     * Maximum cost of travelling to a node.  Dunno why 16 is a
     * good number.
     */
    public final static int INFINITY = 16;
    
    /**
     * GC pool; unavailable for retrieving routes from.
     * <p>
     * DynamicRouteEntries are added to the GC pool when they "expire".
     * Later, they will be deleted from the GC pool, for good.
     * DynamicRouteEntries might get "resuscitated" from the pool.
     * <p>
     * Maps Integer() wrapped destintion addresses to RouteEntries.
     * XXX this is actually another RouteTable instance...
     */
    private final Hashtable gcPool;

    public DynamicRouteTable(RouteTable baseRouteTable)
    {
	RouteEntry res[];
	int lpc;
	
	this.gcPool = new Hashtable();
	this.baseRouteTable = baseRouteTable;

	res = this.baseRouteTable.getRoutes();
	for( lpc = 0; lpc < res.length; lpc++ )
	{
	    /*
	     * This is a bit gross.  We read all the entries
	     * out of the route table and replace them with
	     * DynamicRoute entries.
	     */
	    RouteEntry re = res[lpc];
	    DynamicRouteEntry dre = new DynamicRouteEntry(re);
	    
	    this.baseRouteTable.removeRoute(re.getAddress());
	    dre.setCost(DynamicRouteTable.INFINITY);
	    
	    /*
	     * XXX expire timer will be initialized by
	     * DynamicRouteApp.initEvents()
	     */
	    
	    this.addRoute(dre);
	}
    }
    
    /**
     * Retrive the current route to get to the given active address.
     */
    DynamicRouteEntry getRoute(int activeDest)
    {
	return (DynamicRouteEntry)baseRouteTable.getRoute(activeDest);
    }

    DynamicRouteEntry[] getRoutes()
    {
	DynamicRouteEntry retval[];
	
	synchronized( this.baseRouteTable )
	{
	    retval = new DynamicRouteEntry[this.baseRouteTable.
					  getRouteCount()];
	    this.baseRouteTable.getEntries(retval);
	}
	return retval;
    }
    
    Hashtable gcPool()
    {
	return gcPool;
    }
    
    /**
     * Add a route to the table.
     */
    public void addRoute(DynamicRouteEntry re)
    {
	baseRouteTable.addRoute(re);
    }
    
    /**
     * Remove a route from the table.
     */
    public void removeRoute(DynamicRouteEntry re)
    {
	baseRouteTable.removeRoute(re.getAddress());
    }
}
