/* 
 * MulticastCapsule.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package apps.multicast;

import ants.core.*;

/**
 * Multicast capsule processing.  These correspond to the multicast data
 * packets.
 *
 * @author David Wetherall
 */
public class MulticastCapsule
    extends Capsule
{
    final private static byte[] MID
	= findMID("apps.multicast"
		  + ".MulticastCapsule");
    protected byte[] mid() { return MID; }
    
    final private static byte[] PID
	= findPID("apps.multicast"
		  + ".MulticastCapsule");
    protected byte[] pid() { return PID; }
    
    int group;	// The multicast group (address).
    short destport;	// Destination port
    ByteArray data;	// The actual data for this capsule.
    
    /**
     * Return the length of this capsule.
     */
    public int length() {
	int s = Xdr.INT + Xdr.SHORT + Xdr.BYTEARRAY(data);
	return super.length() + s;
    }
    
    
    /**
     * Encode this capsule.
     */
    public Xdr encode() {
	Xdr xdr = super.encode();
	
	xdr.PUT(group);
	xdr.PUT(destport);
	xdr.PUT(data);
	
	return xdr;
    }
    
    /**
     * Decode this capsule.
     */
    public Xdr decode() {
	Xdr xdr = super.decode();
	
	group = xdr.INT();
	destport = xdr.SHORT();
	data = xdr.BYTEARRAY();
	
	return xdr;
    }
    
    /**
     * Process this capsule at the node.
     *
     * @param n The current node.
     */
    public boolean evaluate(Node n) {
	// Create a key and try to get our soft state.
	MulticastCapsuleCacheKey mkey
	    = new MulticastCapsuleCacheKey(group, getDst());
	MulticastCapsuleCacheValue mval
	    = (MulticastCapsuleCacheValue)n.getCache().get(mkey);
	n.log(0, "MulticastCapsule: Trying to get state for group "
	      + NodeAddress.toString(group)
	      + ", dest " + NodeAddress.toString(getDst())
	      + ", key " + mkey.hashCode());
	
	if (mval != null) {
	    /* We found an entry. */
	    if (mval.addrs != null) {
				/* Send the capsule to all nodes listed. */
		for (int i = 0; i < mval.addrs.length; i++)
		    n.routeForNode(this, mval.addrs[i]);
	    } else {
				/* Deliver to application. */
		n.deliverToApp(this, destport);
	    }
	} else {
	    /* There is no entry, we can only drop it. */
	    int count = ((int)data.getByte(4)) << 24
		| ((int)data.getByte(5)) << 16
		| ((int)data.getByte(6)) << 8
		| (data.getByte(7) & 0xff);
	    n.log(0, "MulticastCapsule: Dropping capsule #" + count
		  + " -- Nowhere to send!");
	}
	return true;
    }
    
    /* Constructors. */
    public MulticastCapsule() { }
    public MulticastCapsule(int group, int target, short destport,
			    ByteArray data) {
	this.group = group;
	this.destport = destport;
	this.data = data;
	setDst(target);
    }
}
